from typing import Union, List

from .workers_cheapening import *  # noqa
from ..base import OptionsGroup
from ..typehints import Strlist
from ..utils import listify


class MuleFarm:
    """Represents a mule farm."""

    def __init__(self, name: str, mule_numbers: Union[int, List[int]]):
        """
        :param name: Farm alias.

        :param mule_numbers: Total mules on farm count,
            or a list of mule numbers.

        """
        # todo http://uwsgi-docs.readthedocs.io/en/latest/Signals.html#signals-targets
        self.name = name
        self.mule_numbers = mule_numbers

    def __str__(self):
        return f"{self.name}:{','.join(map(str, self.mule_numbers))}"


class Workers(OptionsGroup):
    """Workers aka [working] processes."""

    mule_farm = MuleFarm

    def set_basic_params(
            self,
            *,
            count: int = None,
            touch_reload: Strlist = None,
            touch_chain_reload: Strlist = None,
            zombie_reaper: bool = None,
            limit_addr_space: int = None,
            limit_count: int = None,
            cpu_affinity: int = None
    ):
        """

        :param count: Spawn the specified number of workers (processes).
            Set the number of workers for preforking mode.
            This is the base for easy and safe concurrency in your app.
            More workers you add, more concurrent requests you can manage.

            Each worker corresponds to a system process, so it consumes memory,
            choose carefully the right number. You can easily drop your system
            to its knees by setting a too high value.

            Setting ``workers`` to a ridiculously high number will **not**
            magically make your application web scale - quite the contrary.

        :param touch_reload: Trigger reload of (and only) workers
            if the specified file is modified/touched.

        :param touch_chain_reload: Trigger chain workers reload on file touch.
            When in lazy/lazy_apps mode, you can simply destroy a worker to force it to reload the application code.
            A new reloading system named "chain reload", allows you to reload one worker at time
            (opposed to the standard way where all of the workers are destroyed in bulk)

            * http://uwsgi-docs.readthedocs.io/en/latest/articles/TheArtOfGracefulReloading.html#chain-reloading-lazy-apps

        :param zombie_reaper: Call waitpid(-1,...) after each request to get rid of zombies.
            Enables reaper mode. After each request the server will call ``waitpid(-1)``
            to get rid of zombie processes.
            If you spawn subprocesses in your app and you happen to end up with zombie processes
            all over the place you can enable this option. (It really would be better
            if you could fix your application's process spawning usage though.)

        :param limit_addr_space: Limit process address space (vsz) (in megabytes)
            Limits the address space usage of each uWSGI (worker) process using POSIX/UNIX ``setrlimit()``.
            For example, ``limit-as 256`` will disallow uWSGI processes to grow over 256MB of address space.
            Address space is the virtual memory a process has access to. It does *not* correspond to physical memory.
            Read and understand this page before enabling this option: http://en.wikipedia.org/wiki/Virtual_memory

        :param limit_count: Limit the number of spawnable processes.

        :param cpu_affinity: number of cores for each worker (Linux only)
            Set the number of cores (CPUs) to allocate to each worker process.

            * 4 workers, 4 CPUs, affinity is 1,
                each worker is allocated one CPU.

            * 4 workers, 2 CPUs, affinity is 1,
                workers get one CPU each (0; 1; 0; 1).

            * 4 workers, 4 CPUs, affinity is 2,
                workers get two CPUs each in a round-robin fashion (0, 1; 2, 3; 0, 1; 2; 3).

            * 8 workers, 4 CPUs, affinity is 3,
                workers get three CPUs each in a round-robin fashion
                (0, 1, 2; 3, 0, 1; 2, 3, 0; 1, 2, 3; 0, 1, 2; 3, 0, 1; 2, 3, 0; 1, 2, 3).

        """
        # todo map socket to a worker - map-socket = 0:1
        # networking.register_socket - workers_binding

        self.set_count_auto(count)

        self._set('touch-workers-reload', touch_reload, multi=True)
        self._set('touch-chain-reload', touch_chain_reload, multi=True)

        self._set('reaper', zombie_reaper, cast=bool)

        self._set('limit-as', limit_addr_space)
        self._set('limit-nproc', limit_count)

        self._set('cpu-affinity', cpu_affinity)

        return self._section

    def run_command_as_worker(self, command: str, *, after_post_fork_hook: bool = False):
        """Run the specified command as worker.

        :param command:

        :param after_post_fork_hook: Whether to run it after `post_fork` hook.

        """
        self._set('worker-exec2' if after_post_fork_hook else 'worker-exec', command, multi=True)

        return self._section

    def set_count_auto(self, count: int = None):
        """Sets workers count.

        By default sets it to detected number of available cores

        :param count:

        """
        count = count or self._section.vars.CPU_CORES

        self._set('workers', count)

        return self._section

    def set_thread_params(
            self,
            enable: bool = None,
            *,
            count: int = None,
            count_offload: int = None,
            stack_size: int = None,
            no_wait: bool = None
    ):
        """Sets threads related params.

        :param enable: Enable threads in the embedded languages.
            This will allow to spawn threads in your app.

            .. warning:: Threads will simply *not work* if this option is not enabled.
                There will likely be no error, just no execution of your thread code.

        :param count: Run each worker in prethreaded mode with the specified number
            of threads per worker.

            .. warning:: Do not use with ``gevent``.

            .. note:: Enables threads automatically.

        :param count_offload: Set the number of threads (per-worker) to spawn
            for offloading. Default: 0.

            These threads run such tasks in a non-blocking/evented way allowing
            for a huge amount of concurrency. Various components of the uWSGI stack
            are offload-friendly.

            .. note:: Try to set it to the number of CPU cores to take advantage of SMP.

            * http://uwsgi-docs.readthedocs.io/en/latest/OffloadSubsystem.html

        :param stack_size: Set threads stacksize.

        :param no_wait: Do not wait for threads cancellation on quit/reload.

        """
        self._set('enable-threads', enable, cast=bool)
        self._set('no-threads-wait', no_wait, cast=bool)
        self._set('threads', count)
        self._set('offload-threads', count_offload)

        if count:
            self._section.print_out(f'Threads per worker: {count}')

        self._set('threads-stacksize', stack_size)

        return self._section

    def set_mules_params(
            self,
            mules: Union[int, List[int]] = None,
            *,
            touch_reload: Strlist = None,
            harakiri_timeout: int = None,
            farms: List[MuleFarm] = None,
            reload_mercy: int = None,
            msg_buffer: int = None,
            msg_buffer_recv: int = None
    ):
        """Sets mules related params.

        http://uwsgi.readthedocs.io/en/latest/Mules.html

        Mules are worker processes living in the uWSGI stack but not reachable via socket connections,
        that can be used as a generic subsystem to offload tasks.

        :param mules: Add the specified mules or number of mules.

        :param touch_reload: Reload mules if the specified file is modified/touched.

        :param harakiri_timeout: Set harakiri timeout for mule tasks.

        :param farms: Mule farms list.

            Examples:
                * cls_mule_farm('first', 2)
                * cls_mule_farm('first', [4, 5])

        :param reload_mercy: Set the maximum time (in seconds) a mule can take
            to reload/shutdown. Default: 60.

        :param msg_buffer: Set mule message buffer size (bytes) given
            for mule message queue.

        :param msg_buffer: Set mule message recv buffer size (bytes).

        """
        farms = farms or []

        next_mule_number = 1
        farm_mules_count = 0

        for farm in farms:

            if isinstance(farm.mule_numbers, int):
                farm.mule_numbers = list(range(next_mule_number, next_mule_number + farm.mule_numbers))
                next_mule_number = farm.mule_numbers[-1] + 1

            farm_mules_count += len(farm.mule_numbers)

            self._set('farm', farm, multi=True)

        if mules is None and farm_mules_count:
            mules = farm_mules_count

        if isinstance(mules, int):
            self._set('mules', mules)

        elif isinstance(mules, list):

            for mule in mules:
                self._set('mule', mule, multi=True)

        self._set('touch-mules-reload', touch_reload, multi=True)

        self._set('mule-harakiri', harakiri_timeout)
        self._set('mule-reload-mercy', reload_mercy)

        self._set('mule-msg-size', msg_buffer)
        self._set('mule-msg-recv-size', msg_buffer_recv)

        return self._section

    def set_reload_params(
            self,
            *,
            min_lifetime: int = None,
            max_lifetime: int = None,
            max_requests: int = None,
            max_requests_delta: int = None,
            max_addr_space: int = None,
            max_rss: int = None,
            max_uss: int = None,
            max_pss: int = None,
            max_addr_space_forced: int = None,
            max_rss_forced: int = None,
            watch_interval_forced: int = None,
            mercy: int = None
    ):
        """Sets workers reload parameters.

        :param min_lifetime: A worker cannot be destroyed/reloaded unless it has been alive
            for N seconds (default 60). This is an anti-fork-bomb measure.
            Since 1.9

        :param max_lifetime: Reload workers after this many seconds. Disabled by default.
            Since 1.9

        :param max_requests: Reload workers after the specified amount of managed
            requests (avoid memory leaks).
            When a worker reaches this number of requests it will get recycled (killed and restarted).
            You can use this option to "dumb fight" memory leaks.

            Also take a look at the ``reload-on-as`` and ``reload-on-rss`` options
            as they are more useful for memory leaks.

            .. warning:: The default min-worker-lifetime 60 seconds takes priority over `max-requests`.

            Do not use with benchmarking as you'll get stalls
            such as `worker respawning too fast !!! i have to sleep a bit (2 seconds)...`

        :param max_requests_delta: Add (worker_id * delta) to the max_requests value of each worker.

        :param max_addr_space: Reload a worker if its address space usage is higher
            than the specified value in megabytes.

        :param max_rss: Reload a worker if its physical unshared memory (resident set size) is higher
            than the specified value (in megabytes).

        :param max_uss: Reload a worker if Unique Set Size is higher
            than the specified value in megabytes.

            .. note:: Linux only.

        :param max_pss: Reload a worker if Proportional Set Size is higher
            than the specified value in megabytes.

            .. note:: Linux only.

        :param max_addr_space_forced: Force the master to reload a worker if its address space is higher
            than specified megabytes (in megabytes).

        :param max_rss_forced: Force the master to reload a worker
            if its resident set size memory is higher than specified in megabytes.

        :param watch_interval_forced: The memory collector [per-worker] thread memeory watch
            interval (seconds) used for forced reloads. Default: 3.

        :param mercy: Set the maximum time (in seconds) a worker can take
            before reload/shutdown. Default: 60.

        """
        self._set('max-requests', max_requests)
        self._set('max-requests-delta', max_requests_delta)

        self._set('min-worker-lifetime', min_lifetime)
        self._set('max-worker-lifetime', max_lifetime)

        self._set('reload-on-as', max_addr_space)
        self._set('reload-on-rss', max_rss)
        self._set('reload-on-uss', max_uss)
        self._set('reload-on-pss', max_pss)

        self._set('evil-reload-on-as', max_addr_space_forced)
        self._set('evil-reload-on-rss', max_rss_forced)
        self._set('mem-collector-freq', watch_interval_forced)

        self._set('worker-reload-mercy', mercy)

        return self._section

    def set_reload_on_exception_params(
            self,
            *,
            do_reload: bool = None,
            etype: str = None,
            evalue: str = None,
            erepr: str = None
    ):
        """Sets workers reload on exceptions parameters.

        :param do_reload: Reload a worker when an exception is raised.

        :param etype: Reload a worker when a specific exception type is raised.

        :param evalue: Reload a worker when a specific exception value is raised.

        :param erepr: Reload a worker when a specific exception type+value (language-specific) is raised.

        """
        self._set('reload-on-exception', do_reload, cast=bool)
        self._set('reload-on-exception-type', etype)
        self._set('reload-on-exception-value', evalue)
        self._set('reload-on-exception-repr', erepr)

        return self._section

    def set_harakiri_params(self, *, timeout: int = None, verbose: bool = None, disable_for_arh: bool = None):
        """Sets workers harakiri parameters.

        :param timeout: Harakiri timeout in seconds.
            Every request that will take longer than the seconds specified
            in the harakiri timeout will be dropped and the corresponding
            worker is thereafter recycled.

        :param verbose: Harakiri verbose mode.
            When a request is killed by Harakiri you will get a message in the uWSGI log.
            Enabling this option will print additional info (for example,
            the current syscall will be reported on Linux platforms).

        :param disable_for_arh: Disallow Harakiri killings during after-request hook methods.

        """
        self._set('harakiri', timeout)
        self._set('harakiri-verbose', verbose, cast=bool)
        self._set('harakiri-no-arh', disable_for_arh, cast=bool)

        return self._section

    def set_zerg_server_params(self, socket: str, *, clients_socket_pool: Strlist = None):
        """Zerg mode. Zerg server params.

        When your site load is variable, it would be nice to be able to add
        workers dynamically. Enabling Zerg mode you can allow zerg clients to attach
        to your already running server and help it in the work.

        * http://uwsgi-docs.readthedocs.io/en/latest/Zerg.html

        :param socket: Unix socket to bind server to.

            Examples:
                * unix socket - ``/var/run/mutalisk``
                * Linux abstract namespace - ``@nydus``

        :param clients_socket_pool: This enables Zerg Pools.

            .. note:: Expects master process.

            Accepts sockets that will be mapped to Zerg socket.

            * http://uwsgi-docs.readthedocs.io/en/latest/Zerg.html#zerg-pools

        """
        if clients_socket_pool:
            self._set('zergpool', f"{socket}:{','.join(listify(clients_socket_pool))}", multi=True)

        else:
            self._set('zerg-server', socket)

        return self._section

    def set_zerg_client_params(self, server_sockets: Strlist, *, use_fallback_socket: bool = None):
        """Zerg mode. Zergs params.

        :param server_sockets: Attaches zerg to a zerg server.

        :param use_fallback_socket: Fallback to normal sockets if the zerg server is not available

        """
        self._set('zerg', server_sockets, multi=True)

        if use_fallback_socket is not None:
            self._set('zerg-fallback', use_fallback_socket, cast=bool)

            for socket in listify(server_sockets):
                self._section.networking.register_socket(self._section.networking.sockets.default(socket))

        return self._section



import sys
from pathlib import Path
from typing import Union

from ..base import OptionsGroup
from ..exceptions import ConfigurationError
from ..typehints import Strint, Strlist

AUTO = -1


class Python(OptionsGroup):
    """Python plugin options.

    .. note:: By default the plugin does not initialize the GIL.
        This means your app-generated threads will not run.
        If you need threads, remember to enable them with ``enable_threads``.

    """
    plugin = True

    def set_basic_params(
            self,
            *,
            version: Strint = AUTO,
            python_home: str = None,
            enable_threads: bool = None,
            search_path: str = None,
            python_binary: str = None,
            tracebacker_path: str = None,
            plugin_dir: str = None,
            os_env_reload: bool = None,
            optimization_level: int = None
    ):
        """

        :param version: Python version plugin supports.

            Example:
                * 3 - version 3
                * <empty> - version 2
                * <default> - version deduced by uwsgiconf

        :param python_home: Set python executable directory - PYTHONHOME/virtualenv.

        :param bool enable_threads: Enable threads in the embedded languages.
            This will allow to spawn threads in your app.

            .. warning:: Threads will simply *not work* if this option is not enabled.
                         There will likely be no error, just no execution of your thread code.

        :param search_path: Add directory (or an .egg or a glob) to the Python search path.

            .. note:: This can be specified up to 64 times.

        :param python_binary: Set python program name.

        :param tracebacker_path: Enable the uWSGI Python tracebacker.
            http://uwsgi-docs.readthedocs.io/en/latest/Tracebacker.html

        :param plugin_dir: Directory to search for plugin.

        :param os_env_reload: Force ``os.environ`` reloading for every request.
            Used to allow setting of ``UWSGI_SETENV`` for Python applications.

        :param optimization_level: Python optimization level (see ``-O`` argument).
            .. warning:: This may be dangerous for some apps.

        """
        self._set_name(version)

        self._section.workers.set_thread_params(enable=enable_threads)
        self._set('py-tracebacker', tracebacker_path)
        self._set('py-program-name', python_binary)
        self._set('pyhome', python_home)
        self._set('pythonpath', search_path, multi=True)
        self._set('reload-os-env', os_env_reload, cast=bool)
        self._set('optimize', optimization_level)

        self._section.set_plugins_params(search_dirs=plugin_dir)

        return self._section

    def _set_name(self, version: Strint = AUTO):
        """Returns plugin name."""

        name = 'python'

        if version:
            if version is AUTO:
                version = sys.version_info[0]

                if version == 2:
                    version = ''

            name = f'{name}{version}'

        self.name = name

    def _get_name(self, *args, **kwargs):
        name = self.name

        if not name:
            self._set_name()

        return self.name

    def set_app_args(self, *args):
        """Sets ``sys.argv`` for python apps.

        Examples:
            * pyargv="one two three" will set ``sys.argv`` to ``('one', 'two', 'three')``.

        :param args:

        """
        if args:
            self._set('pyargv', ' '.join(args))

        return self._section

    def set_wsgi_params(self, *, module: Union[str, Path] = None, callable_name: str = None, env_strategy: str = None):
        """Set wsgi related parameters.

        :param module:
            * load .wsgi file as the Python application
            * load a WSGI module as the application.

            .. note:: The module (sans ``.py``) must be importable, ie. be in ``PYTHONPATH``.

            Examples:
                * mypackage.my_wsgi_module -- read from `application` attr of mypackage/my_wsgi_module.py
                * mypackage.my_wsgi_module:my_app -- read from `my_app` attr of mypackage/my_wsgi_module.py

        :param callable_name: Set WSGI callable name. Default: application.

        :param env_strategy: Strategy for allocating/deallocating
            the WSGI env, can be:

            * ``cheat`` - preallocates the env dictionary on uWSGI startup and clears it
                after each request. Default behaviour for uWSGI <= 2.0.x

            * ``holy`` - creates and destroys the environ dictionary at each request.
                Default behaviour for uWSGI >= 2.1

        """
        module = str(module or '')

        if '/' in module:
            self._set('wsgi-file', module, condition=module)

        else:
            self._set('wsgi', module, condition=module)

        self._set('callable', callable_name)
        self._set('wsgi-env-behaviour', env_strategy)

        return self._section

    def eval_wsgi_entrypoint(self, code: str):
        """Evaluates Python code as WSGI entry point.

        :param code:

        """
        self._set('eval', code)

        return self._section

    def set_autoreload_params(self, *, scan_interval: int = None, ignore_modules: Strlist = None):
        """Sets autoreload related parameters.

        :param scan_interval: Seconds. Monitor Python modules' modification times to trigger reload.

            .. warning:: Use only in development.

        :param ignore_modules: Ignore the specified module during auto-reload scan.

        """
        self._set('py-auto-reload', scan_interval)
        self._set('py-auto-reload-ignore', ignore_modules, multi=True)

        return self._section

    def register_module_alias(self, alias: str, module_path: str, *, after_init: bool = False):
        """Adds an alias for a module.

        http://uwsgi-docs.readthedocs.io/en/latest/PythonModuleAlias.html

        :param alias:
        :param module_path:
        :param after_init: add a python module alias after uwsgi module initialization

        """
        command = 'post-pymodule-alias' if after_init else 'pymodule-alias'
        self._set(command, f'{alias}={module_path}', multi=True)

        return self._section

    def import_module(self, modules: Strint, *, shared: bool = False, into_spooler: bool = False):
        """Imports a python module.

        :param modules:

        :param shared: If shared import is done once in master process.
            Otherwise import a python module in all of the processes.
            This is done after fork but before request processing.

        :param into_spooler: Import a python module in the spooler.
            http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html

        """
        if all((shared, into_spooler)):
            raise ConfigurationError('Unable to set both `shared` and `into_spooler` flags')

        if into_spooler:
            command = 'spooler-python-import'
        else:
            command = 'shared-python-import' if shared else 'python-import'

        self._set(command, modules, multi=True)

        return self._section

    def run_module(self, module: str):
        """Runs a Python script in the uWSGI environment.

        :param module:

        """
        self._set('pyrun', module)

        return self._section



from collections import namedtuple
from typing import List, Callable

from .. import uwsgi
from ..exceptions import UwsgiconfException
from ..settings import get_maintenance_inplace
from ..utils import get_logger

_LOG = get_logger(__name__)


SignalDescription = namedtuple('SignalDescription', ['num', 'target', 'func'])
"""Registered signal information."""

registry_signals: List[SignalDescription] = []
"""Registered signals."""


def get_available_num() -> int:
    """Returns first available signal number.

    :raises UwsgiconfException: If no signal is available.

    """
    for signum in range(0, 256):
        if not uwsgi.signal_registered(signum):
            return signum

    raise UwsgiconfException('No uWSGI signals available.')


def get_last_received() -> 'Signal':
    """Get the last signal received."""
    return Signal(uwsgi.signal_received())


class Signal:
    """Represents uWSGI signal.

    .. warning:: If you define a new function in worker1 and register
        it as a signal handler, only worker1 can run it. The best way to register signals
        is defining them in the master (.runtime.uwsgi.postfork_hooks.add), so all workers see them.

    .. code-block:: python

        signal = Signal()

        @signal.register_handler()
        def somefunc():
            pass

        # or the same:

        @signal
        def somefunc():
            pass

    """
    __slots__ = ['num']

    def __init__(self, num: int = None):
        """
        :param int num: Signal number (0-255).

            .. note:: If not set it will be chosen automatically.

        """
        self.num = num or get_available_num()

    def __int__(self):
        return self.num

    def __call__(self, func: Callable):
        # Allows using object as a decorator.
        return self.register_handler()(func)

    @property
    def registered(self) -> bool:
        """Whether the signal is registered."""
        return uwsgi.signal_registered(self.num) or False

    def register_handler(self, *, target: str = None) -> Callable:
        """Decorator for a function to be used as a signal handler.

        .. code-block:: python

            signal = Signal()

            @signal.register_handler()
            def somefunc():
                pass

        :param target: Where this signal will be delivered to. Default: ``worker``.

            * ``workers``  - run the signal handler on all the workers
            * ``workerN`` - run the signal handler only on worker N
            * ``worker``/``worker0`` - run the signal handler on the first available worker
            * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

            * ``mules`` - run the signal handler on all of the mules
            * ``muleN`` - run the signal handler on mule N
            * ``mule``/``mule0`` - run the signal handler on the first available mule

            * ``spooler`` - run the signal on the first available spooler
            * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

            * http://uwsgi.readthedocs.io/en/latest/Signals.html#signals-targets

        """
        target = target or 'worker'
        sign_num = self.num

        def wrapper(func: Callable):

            _LOG.debug(f"Registering '{func.__name__}' as signal '{sign_num}' handler ...")

            uwsgi.register_signal(sign_num, target, func)
            registry_signals.append(SignalDescription(sign_num, target, func))

            return func

        return wrapper

    def send(self, *, remote: str = None):
        """Sends the signal to master or remote.

        When you send a signal, it is copied into the master's queue.
        The master will then check the signal table and dispatch the messages.

        :param remote: Remote address.

        :raises ValueError: If remote rejected the signal.

        :raises OSError: If unable to deliver to remote.

        """
        uwsgi.signal(self.num, *([remote] if remote is not None else []))

    def wait(self):
        """Waits for the given of any signal.

        Block the process/thread/async core until a signal is received. Use signal_received to get the number of
        the signal received. If a registered handler handles a signal, signal_wait will be interrupted and the actual
        handler will handle the signal.

        * http://uwsgi-docs.readthedocs.io/en/latest/Signals.html#signal-wait-and-signal-received

        :raises SystemError: If something went wrong.

        """
        uwsgi.signal_wait(self.num)


def _automate_signal(target, func):

    if get_maintenance_inplace():
        # Prevent background works in maintenance mode.
        return lambda *args, **kwarg: None

    if target is None or isinstance(target, str):
        sig = Signal()

        func(sig)

        return sig.register_handler(target=target)

    func(target)



from typing import Dict

from .. import uwsgi


class _Request:
    """Current request information."""

    __slots__ = []

    @property
    def env(self) -> Dict[str, str]:
        """Request environment dictionary."""
        return uwsgi.env

    @property
    def id(self) -> int:
        """Returns current request number (handled by worker on core)."""
        return uwsgi.request_id()

    @property
    def total_count(self) -> int:
        """Returns the total number of requests managed so far by the pool of uWSGI workers."""
        return uwsgi.total_requests()

    @property
    def fd(self) -> int:
        """Returns current request file descriptor."""
        return uwsgi.connection_fd()

    @property
    def content_length(self) -> int:
        """Returns current post content length."""
        return uwsgi.cl()

    def log(self):
        """Instructs uWSGI to log current request data."""
        uwsgi.log_this_request()

    def add_var(self, name: str, value: str) -> bool:
        """Registers custom request variable.

        Can be used for better integration with the internal routing subsystem.

        :param name:

        :param value:

        :raises ValueError: If buffer size is not enough.

        """
        return uwsgi.add_var(name, value)



from ..base import OptionsGroup, TemplatedValue
from .logging_loggers import *
from .logging_encoders import *
from ..utils import listify


class Var(TemplatedValue):

    def __str__(self):
        return '%(' + super().__str__() + ')'


class VarMetric(Var):

    tpl = 'metric.%s'


class VarRequestVar(Var):

    tpl = 'var.%s'

    def __init__(self, name):
        name = name.replace('-', '_').upper()  # Normalize.
        super().__init__(name)


class Logging(OptionsGroup):
    """Logging.

    * http://uwsgi.readthedocs.io/en/latest/Logging.html
    * http://uwsgi-docs.readthedocs.io/en/latest/LogFormat.html

    """

    class loggers:
        """Loggers available for ``add_logger()``."""

        file = LoggerFile
        fd = LoggerFileDescriptor
        stdio = LoggerStdIO
        mongo = LoggerMongo
        redis = LoggerRedis
        socket = LoggerSocket
        syslog = LoggerSyslog
        rsyslog = LoggerRsyslog
        zeromq = LoggerZeroMq

        # todo consider adding other loggers: crypto, graylog2, systemd

    class encoders:
        """Loggers available for ``add_logger_encoder()``."""

        compress = EncoderCompress
        format = EncoderFormat
        gzip = EncoderGzip
        json = EncoderJson
        newline = EncoderNewline
        prefix = EncoderPrefix
        suffix = EncoderSuffix

        # todo consider adding msgpack encoder

    def set_basic_params(
            self, *, no_requests=None, template=None, memory_report=None, prefix=None, prefix_date=None,
            apply_strftime=None, response_ms=None, ip_x_forwarded=None):
        """

        :param bool no_requests: Disable requests logging - only uWSGI internal messages
            and errors will be logged.

        :param str template: Set advanced format for request logging.
            This template string can use variables from ``Logging.Vars``.

        :param str prefix: Prefix log items with a string.

        :param str|bool prefix_date: Prefix log items with date string.
        
            .. note:: This can be ``True`` or contain formatting placeholders (e.g. %Y-%m-%d %H:%M:%S) 
               if used with ``apply_strftime``.

        :param int memory_report: Enable memory report.
                * **1** - basic (default);
                * **2** - uss/pss (Linux only)

        :param bool apply_strftime: Apply strftime to dates in log entries. 
            E.g. ``prefix_date`` can contain format placeholders. See also ``vars.REQ_START_FORMATTED``.

        :param bool response_ms: Report response time in microseconds instead of milliseconds.

        :param bool ip_x_forwarded: Use the IP from X-Forwarded-For header instead of REMOTE_ADDR.
            Used when uWSGI is run behind multiple proxies.

        """
        self._set('disable-logging', no_requests, cast=bool)
        self._set('log-format', template)
        self._set('memory-report', memory_report)
        self._set('log-prefix', prefix)

        if isinstance(prefix_date, str) and '%' in prefix_date:
            prefix_date = prefix_date.replace('%', '%%')  # Escaping.
                
        self._set('log-date', prefix_date, cast=(bool if isinstance(prefix_date, bool) else None))

        self._set('logformat-strftime', apply_strftime, cast=bool)
        self._set('log-micros', response_ms, cast=bool)
        self._set('log-x-forwarded-for', ip_x_forwarded, cast=bool)

        return self._section

    def log_into(self, target, *, before_priv_drop=True):
        """Simple file or UDP logging.

        .. note:: This doesn't require any Logger plugin and can be used
            if no log routing is required.

        :param str target: Filepath or UDP address.

        :param bool before_priv_drop: Whether to log data before or after privileges drop.

        """
        command = 'logto'

        if not before_priv_drop:
            command += '2'

        self._set(command, target)

        return self._section

    def set_file_params(
            self, *, reopen_on_reload=None, trucate_on_statup=None, max_size=None, rotation_fname=None,
            touch_reopen=None, touch_rotate=None, owner=None, mode=None):
        """Set various parameters related to file logging.

        :param bool reopen_on_reload: Reopen log after reload.

        :param bool trucate_on_statup: Truncate log on startup.

        :param int max_size: Set maximum logfile size in bytes after which log should be rotated.

        :param str rotation_fname: Set log file name after rotation.

        :param str|list touch_reopen: Trigger log reopen if the specified file
            is modified/touched.

            .. note:: This can be set to a file touched by ``postrotate`` script of ``logrotate``
                to implement rotation.

        :param str|list touch_rotate: Trigger log rotation if the specified file
            is modified/touched.

        :param str owner: Set owner chown() for logs.
        
        :param str mode: Set mode chmod() for logs.

        """
        self._set('log-reopen', reopen_on_reload, cast=bool)
        self._set('log-truncate', trucate_on_statup, cast=bool)
        self._set('log-maxsize', max_size)
        self._set('log-backupname', rotation_fname)

        self._set('touch-logreopen', touch_reopen, multi=True)
        self._set('touch-logrotate', touch_rotate, multi=True)

        self._set('logfile-chown', owner)
        self._set('logfile-chmod', mode)

        return self._section

    def set_filters(self, *, include=None, exclude=None, write_errors=None, write_errors_tolerance=None, sigpipe=None):
        """Set various log data filters.

        :param str|list include: Show only log lines matching the specified regexp.

            .. note:: Requires enabled PCRE support.

        :param str|list exclude: Do not show log lines matching the specified regexp.

            .. note:: Requires enabled PCRE support.

        :param bool write_errors: Log (annoying) write()/writev() errors. Default: ``True``.

            .. note:: If both this and ``sigpipe`` set to ``False``, it's the same
               as setting ``write-errors-exception-only`` uWSGI option.

        :param int write_errors_tolerance: Set the maximum number of allowed write errors before exception
            is raised. Default: no tolerance.

            .. note:: Available for Python, Perl, PHP.

        :param bool sigpipe: Log (annoying) SIGPIPE. Default: ``True``.

            .. note:: If both this and ``write_errors`` set to ``False``, it's the same
               as setting ``write-errors-exception-only`` uWSGI option.

        """
        if write_errors is not None:
            self._set('ignore-write-errors', not write_errors, cast=bool)

        if sigpipe is not None:
            self._set('ignore-sigpipe', not sigpipe, cast=bool)

        self._set('write-errors-tolerance', write_errors_tolerance)

        for line in listify(include):
            self._set('log-filter', line, multi=True)

        for line in listify(exclude):
            self._set('log-drain', line, multi=True)

        return self._section

    def set_requests_filters(
            self, *, slower=None, bigger=None, status_4xx=None, status_5xx=None,
            no_body=None, sendfile=None, io_errors=None):
        """Set various log data filters.

        :param int slower: Log requests slower than the specified number of milliseconds.

        :param int bigger: Log requests bigger than the specified size in bytes.

        :param status_4xx: Log requests with a 4xx response.

        :param status_5xx: Log requests with a 5xx response.

        :param bool no_body: Log responses without body.

        :param bool sendfile: Log sendfile requests.

        :param bool io_errors: Log requests with io errors.

        """
        self._set('log-slow', slower)
        self._set('log-big', bigger)
        self._set('log-4xx', status_4xx, cast=bool)
        self._set('log-5xx', status_5xx, cast=bool)
        self._set('log-zero', no_body, cast=bool)
        self._set('log-sendfile', sendfile, cast=bool)
        self._set('log-ioerror', io_errors, cast=bool)

        return self._section

    def set_master_logging_params(
            self, enable=None, *, dedicate_thread=None, buffer=None,
            sock_stream=None, sock_stream_requests_only=None):
        """Sets logging params for delegating logging to master process.

        :param bool enable: Delegate logging to master process.
            Delegate the write of the logs to the master process
            (this will put all of the logging I/O to a single process).
            Useful for system with advanced I/O schedulers/elevators.

        :param bool dedicate_thread: Delegate log writing to a thread.

            As error situations could cause the master to block while writing
            a log line to a remote server, it may be a good idea to use this option and delegate
            writes to a secondary thread.

        :param int buffer: Set the buffer size for the master logger in bytes.
            Bigger log messages will be truncated.

        :param bool|tuple sock_stream: Create the master logpipe as SOCK_STREAM.

        :param bool|tuple sock_stream_requests_only: Create the master requests logpipe as SOCK_STREAM.

        """
        self._set('log-master', enable, cast=bool)
        self._set('threaded-logger', dedicate_thread, cast=bool)
        self._set('log-master-bufsize', buffer)

        self._set('log-master-stream', sock_stream, cast=bool)

        if sock_stream_requests_only:
            self._set('log-master-req-stream', sock_stream_requests_only, cast=bool)

        return self._section

    def print_loggers(self):
        """Print out available (built) loggers."""

        self._set('loggers-list', True, cast=bool)

        return self._section

    def add_logger(self, logger, *, requests_only=False, for_single_worker=False):
        """Set/add a common logger or a request requests only.

        :param str|list|Logger|list[Logger] logger:

        :param bool requests_only: Logger used only for requests information messages.

        :param bool for_single_worker: Logger to be used in single-worker setup.


        """
        if for_single_worker:
            command = 'worker-logger-req' if requests_only else 'worker-logger'
        else:
            command = 'logger-req' if requests_only else 'logger'

        for logger in listify(logger):
            self._set(command, logger, multi=True)

        return self._section

    def add_logger_route(self, logger, matcher, *, requests_only=False):
        """Log to the specified named logger if regexp applied on log item matches.

        :param str|list|Logger|list[Logger] logger: Logger to associate route with.

        :param str matcher: Regular expression to apply to log item.

        :param bool requests_only: Matching should be used only for requests information messages.

        """
        command = 'log-req-route' if requests_only else 'log-route'

        for logger in listify(logger):
            self._set(command, f'{logger} {matcher}', multi=True)

        return self._section

    def add_logger_encoder(self, encoder, *, logger=None, requests_only=False, for_single_worker=False):
        """Add an item in the log encoder or request encoder chain.

        * http://uwsgi-docs.readthedocs.io/en/latest/LogEncoders.html

            .. note:: Encoders automatically enable master log handling (see ``.set_master_logging_params()``).

            .. note:: For best performance consider allocating a thread
                for log sending with ``dedicate_thread``.

        :param str|list|Encoder encoder: Encoder (or a list) to add into processing.

        :param str|Logger logger: Logger apply associate encoders to.

        :param bool requests_only: Encoder to be used only for requests information messages.

        :param bool for_single_worker: Encoder to be used in single-worker setup.

        """
        if for_single_worker:
            command = 'worker-log-req-encoder' if requests_only else 'worker-log-encoder'
        else:
            command = 'log-req-encoder' if requests_only else 'log-encoder'

        for encoder in listify(encoder):

            value = f'{encoder}'

            if logger:
                if isinstance(logger, Logger):
                    logger = logger.alias

                value += f':{logger}'

            self._set(command, value, multi=True)

        return self._section

    class vars:
        """Variables available for custom log formatting."""

        # The following are taken blindly from the internal wsgi_request structure of the current request.

        REQ_URI = '%(uri)'
        """REQUEST_URI from ``wsgi_request`` of the current request."""

        REQ_METHOD = '%(method)'
        """REQUEST_METHOD from ``wsgi_request`` of the current request."""

        REQ_REMOTE_USER = '%(user)'
        """REMOTE_USER from ``wsgi_request`` of the current request."""

        REQ_REMOTE_ADDR = '%(addr)'
        """REMOTE_ADDR from ``wsgi_request`` of the current request."""

        REQ_HTTP_HOST = '%(host)'
        """HTTP_HOST from ``wsgi_request`` of the current request."""

        REQ_SERVER_PROTOCOL = '%(proto)'
        """SERVER_PROTOCOL from ``wsgi_request`` of the current request."""

        REQ_USER_AGENT = '%(uagent)'
        """HTTP_USER_AGENT from ``wsgi_request`` of the current request."""

        REQ_REFERER = '%(referer)'
        """HTTP_REFERER from ``wsgi_request`` of the current request."""

        # The following are simple functions called to generate the logvar value.

        REQ_START_TS = '%(time)'
        """Timestamp of the start of the request. E.g.: 1512623650"""

        REQ_START_CTIME = '%(ctime)'
        """Ctime of the start of the request. E.g.: Thu Dec  7 08:05:35 2017"""

        REQ_START_UNIX_US = '%(tmsecs)'
        """Timestamp of the start of the request in milliseconds since the epoch.

        .. note:: since 1.9.21

        """

        REQ_START_UNIX_MS = '%(tmicros)'
        """Timestamp of the start of the request in microseconds since the epoch.

        .. note:: since 1.9.21

        """

        REQ_START_HUMAN = '%(ltime)'
        """Human-formatted (Apache style) request time."""

        REQ_START_FORMATTED = '%(ftime)'
        """Request time formatted with ``apply_strftime``. 

        .. note:: Use ``apply_strftime`` and placeholders.

        """

        REQ_SIZE_BODY = '%(cl)'
        """Request content body size."""

        REQ_COUNT_VARS_CGI = '%(vars)'
        """Number of CGI vars in the request."""

        REQ_COUNT_ERR_READ = '%(rerr)'
        """Number of read errors for the request.

        .. note:: since 1.9.21

        """

        REQ_COUNT_ERR_WRITE = '%(werr)'
        """Number of write errors for the request.

        .. note:: since 1.9.21

        """

        REQ_COUNT_ERR = '%(ioerr)'
        """Number of write and read errors for the request.

        .. note:: since 1.9.21

        """

        RESP_STATUS = '%(status)'
        """HTTP response status code."""

        RESP_TIME_US = '%(micros)'
        """Response time in microseconds. E.g.: 1512623650704"""

        RESP_TIME_MS = '%(msecs)'
        """Response time in milliseconds. E.g.: 1512623650704413"""

        RESP_SIZE = '%(size)'
        """Response body size + response headers size."""

        RESP_SIZE_HEADERS = '%(hsize)'
        """Response headers size."""

        RESP_SIZE_BODY = '%(rsize)'
        """Response body size."""

        RESP_COUNT_HEADERS = '%(headers)'
        """Number of generated response headers."""

        TIME_UNIX = '%(epoch)'
        """The current time in Unix format."""

        WORKER_PID = '%(pid)'
        """pid of the worker handling the request."""

        WORKER_ID = '%(wid)'
        """id of the worker handling the request."""

        ASYNC_SWITCHES = '%(switches)'
        """Number of async switches."""

        CORE = '%(core)'
        """The core running the request."""

        MEM_VSZ = '%(vsz)'
        """Address space/virtual memory usage (in bytes)."""

        MEM_RSS = '%(rss)'
        """RSS memory usage (in bytes)."""

        MEM_VSZ_MB = '%(vszM)'
        """Address space/virtual memory usage (in megabytes)."""

        MEM_RSS_MV = '%(rssM)'
        """RSS memory usage (in megabytes)."""

        SIZE_PACKET_UWSGI = '%(pktsize)'
        """Size of the internal request uwsgi packet."""

        MOD1 = '%(modifier1)'
        """``modifier1`` of the request. See ``.routing.modifiers``."""

        MOD2 = '%(modifier2)'
        """``modifier2`` of the request. See ``.routing.modifiers``."""

        metric = VarMetric
        """Metric value (see The Metrics subsystem)."""

        request_var = VarRequestVar
        """Request variable value."""



from uwsgiconf.runtime.locking import *


def test_locking():

    @lock()
    def locked():
        pass

    locked()

    with lock():
        pass

    my_lock = lock(10)

    assert int(my_lock) == 10
    assert not my_lock.is_set



from typing import Tuple

from ..config import Section
from ..utils import filter_locals
from ..typehints import Strlist


class Broodlord:
    """This mode is a way for a vassal to ask for reinforcements to the Emperor.

    Reinforcements are new vassals spawned on demand generally bound on the same socket.

    .. warning:: If you are looking for a way to dynamically adapt the number
        of workers of an instance, check the Cheaper subsystem - adaptive process spawning mode.

        *Broodlord mode is for spawning totally new instances.*

    * http://uwsgi-docs.readthedocs.io/en/latest/Broodlord.html

    """
    def __init__(
            self,
            zerg_socket: str,
            *,
            zerg_die_on_idle: int = None,
            vassals_home: Strlist = None,
            zerg_count: int = None,
            vassal_overload_sos_interval: int = None,
            vassal_queue_items_sos: int = None,
            section_emperor: Section = None,
            section_zerg: Section = None
    ):
        """
        :param zerg_socket: Unix socket to bind server to.

        :param zerg_die_on_idle: A number of seconds after which an idle zerg will be destroyed.

        :param vassals_home: Set vassals home.

        :param zerg_count: Maximum number of zergs to spawn.

        :param vassal_overload_sos_interval: Ask emperor for reinforcement when overloaded.
            Accepts the number of seconds to wait between asking for a new reinforcements.

        :param vassal_queue_items_sos: Ask emperor for sos if backlog queue has more
            items than the value specified

        :param section_emperor: Custom section object.

        :param section_zerg: Custom section object.

        """
        self.socket = zerg_socket
        self.vassals_home = vassals_home
        self.die_on_idle = zerg_die_on_idle
        self.broodlord_params = filter_locals(
            locals(), include=[
                'zerg_count',
                'vassal_overload_sos_interval',
                'vassal_queue_items_sos',
            ])

        section_emperor = section_emperor or Section()
        section_zerg = section_zerg or Section.derive_from(section_emperor)

        self.section_emperor = section_emperor
        self.section_zerg = section_zerg

    def configure(self) -> Tuple[Section, Section]:
        """Configures broodlord mode and returns emperor and zerg sections."""

        section_emperor = self.section_emperor
        section_zerg = self.section_zerg

        socket = self.socket

        section_emperor.workers.set_zerg_server_params(socket=socket)
        section_emperor.empire.set_emperor_params(vassals_home=self.vassals_home)
        section_emperor.empire.set_mode_broodlord_params(**self.broodlord_params)

        section_zerg.name = 'zerg'
        section_zerg.workers.set_zerg_client_params(server_sockets=socket)

        if self.die_on_idle:
            section_zerg.master_process.set_idle_params(timeout=30, exit=True)

        return section_emperor, section_zerg



from ..base import OptionsGroup, ParametrizedValue
from ..utils import KeyValue, filter_locals
from .routing_modifiers import Modifier, ModifierWsgi
from .networking_sockets import SocketShared


class RouterBase(OptionsGroup):

    alias = None
    on_command = None

    def _set_aliased(self, *args, **kwargs):
        args = list(args)
        args[0] = f'{self.alias}-{args[0]}'
        self._set(*args, **kwargs)

    def __init__(self, on=None):
        """
        :param SocketShared|str on: Activates the router on the given address.
        """
        self._make_section_like()

        command = self.on_command or self.alias

        self._set(command, on)

        super().__init__()

    def _contribute_to_opts(self, target):
        target_section = target._section
        networking = target_section.networking

        def handle_shared(value):
            # Handle shared sockets as addresses.
            networking.register_socket(value)
            return networking._get_shared_socket_idx(value)

        for key, value in self._section._opts.items():

            if isinstance(value, KeyValue):
                on = value.locals_dict.get('on')

                if isinstance(on, SocketShared):
                    # Replace address with shared socket index.
                    value.locals_dict['on'] = handle_shared(on)

            else:
                if isinstance(value, SocketShared):
                    value = handle_shared(value)

            target_section._opts[key] = value


class _RouterCommon(RouterBase):

    def set_basic_params(
            self, *, workers=None, zerg_server=None, fallback_node=None, concurrent_events=None,
            cheap_mode=None, stats_server=None):
        """

        :param int workers: Number of worker processes to spawn.

        :param str zerg_server: Attach the router to a zerg server.

        :param str fallback_node: Fallback to the specified node in case of error.

        :param int concurrent_events: Set the maximum number of concurrent events router can manage.

            Default: system dependent.

        :param bool cheap_mode: Enables cheap mode. When the router is in cheap mode,
            it will not respond to requests until a node is available.
            This means that when there are no nodes subscribed, only your local app (if any) will respond.
            When all of the nodes go down, the router will return in cheap mode.

        :param str stats_server: Router stats server address to run at.

        """
        self._set_aliased('workers', workers)
        self._set_aliased('zerg', zerg_server)
        self._set_aliased('fallback', fallback_node)
        self._set_aliased('events', concurrent_events)
        self._set_aliased('cheap', cheap_mode, cast=bool)
        self._set_aliased('stats', stats_server)

        return self

    def set_connections_params(self, *, harakiri=None, timeout_socket=None, retry_delay=None):
        """Sets connection-related parameters.

        :param int harakiri: Set gateway harakiri timeout (seconds).

        :param int timeout_socket: Node socket timeout (seconds). Default: 60.

        :param int retry_delay: Retry connections to dead static nodes after the specified
            amount of seconds. Default: 30.

        """
        self._set_aliased('harakiri', harakiri)
        self._set_aliased('timeout', timeout_socket)
        self._set_aliased('gracetime', retry_delay)

        return self


class Forwarder(ParametrizedValue):

    opt_key = True


class ForwarderPath(Forwarder):
    """Use the specified base (allows %s pattern) for mapping requests to UNIX sockets.

    Examples:
        * /tmp/sockets/
        * /tmp/sockets/%s/uwsgi.sock

    * http://uwsgi.readthedocs.io/en/latest/Fastrouter.html#way-1-fastrouter-use-base
    * http://uwsgi.readthedocs.io/en/latest/Fastrouter.html#way-2-fastrouter-use-pattern

    """

    name = 'use-base'

    def __init__(self, sockets_dir):
        """
        :param str sockets_dir: UNIX sockets directory.
            Allows %s to denote key (domain).

        """
        if '%s' in sockets_dir:
            self.name = 'use-pattern'

        super().__init__(sockets_dir)


class ForwarderCode(Forwarder):
    """Forwards requests to nodes returned by a function.

    This allows using user defined functions to calculate.
    Function must accept key (domain).

    * http://uwsgi.readthedocs.io/en/latest/Fastrouter.html#way-5-fastrouter-use-code-string

    .. warning:: Remember to not put blocking code in your functions.
        The router is totally non-blocking, do not ruin it!

    """
    name = 'use-code-string'
    args_joiner = ':'

    def __init__(self, script, func, *, modifier=None):
        """
        :param str script: Script (module for Python) name to get function from.

        :param str func:  Function name.

        :param Modifier modifier: Routing modifier.

        """
        modifier = modifier or ModifierWsgi
        super().__init__(modifier.code, script, func)


class ForwarderCache(Forwarder):
    """Uses uWSGI cache to get target nodes from.

    * http://uwsgi.readthedocs.io/en/latest/Fastrouter.htmlway-3-fastrouter-use-cache

    """
    name = 'use-cache'

    def __init__(self, cache_name=None):
        """
        :param str cache_name: Cache name to use.
        """
        super().__init__(cache_name)


class ForwarderSocket(Forwarder):
    """Forwards request to the specified uwsgi socket."""

    name = 'use-socket'

    def __init__(self, socket):
        """
        :param str socket: Socket filepath.
        """
        super().__init__(socket)


class ForwarderSubscriptionServer(Forwarder):
    """Forwards requests to nodes returned by the subscription server.

    Subscriptions are simple UDP packets that instruct the router which domain maps to which instance or instances.

    To subscribe to such a subscription server use `.subscriptions.subscribe()`.

    * http://uwsgi.readthedocs.io/en/latest/Fastrouter.html#way-4-fastrouter-subscription-server

    """
    name = 'subscription-server'

    def __init__(self, address):
        """
        :param str address: Address (including port) to run the subscription server on.
        """
        super().__init__(address)


class _RouterWithForwarders(_RouterCommon):

    class forwarders:
        """Forwarders are kind of registries containing information
        on where to forward requests.

        """
        cache = ForwarderCache
        code = ForwarderCode
        path = ForwarderPath
        socket = ForwarderSocket
        subscription_server = ForwarderSubscriptionServer

    def __init__(self, on=None, *, forward_to=None):
        """Activates the router on the given address.

        :param SocketShared|str on: Activates the router on the given address.

        :param Forwarder|str|list[str] forward_to: Where to forward requests.
            Expects a forwarder instance or one or more node names.

        """
        super().__init__(on)

        if forward_to is not None:
            if isinstance(forward_to, Forwarder):

                value = f'{forward_to}'
                self._set_aliased(forward_to.name, value, multi=True)

            else:
                self._set_aliased('to', forward_to, multi=True)

    def set_basic_params(
            self, *, workers=None, zerg_server=None, fallback_node=None, concurrent_events=None,
            cheap_mode=None, stats_server=None, quiet=None, buffer_size=None):
        """

        :param int workers: Number of worker processes to spawn.

        :param str zerg_server: Attach the router to a zerg server.

        :param str fallback_node: Fallback to the specified node in case of error.

        :param int concurrent_events: Set the maximum number of concurrent events router can manage.

            Default: system dependent.

        :param bool cheap_mode: Enables cheap mode. When the router is in cheap mode,
            it will not respond to requests until a node is available.
            This means that when there are no nodes subscribed, only your local app (if any) will respond.
            When all of the nodes go down, the router will return in cheap mode.

        :param str stats_server: Router stats server address to run at.

        :param bool quiet: Do not report failed connections to instances.

        :param int buffer_size: Set the internal buffer max size - size of a request (request-body excluded),
            this generally maps to the size of request headers.  Default: 4096 bytes (4k) / page size.

            The amount of variables you can add per-request is limited by the uwsgi packet buffer ().
            You can increase it up to 65535 (64k) with this option.

            .. note: If you receive a bigger request (for example with big cookies or query string)
                so that "invalid request block size" is logged in your logs you may need to increase it.
                It is a security measure too, so adapt to your app needs instead of maxing it out.

        :param forward_to: Forward request to the given node.

        """
        super().set_basic_params(**filter_locals(locals(), drop=[
            'quiet',
            'buffer_size',
        ]))

        self._set_aliased('quiet', quiet, cast=bool)
        self._set_aliased('buffer-size', buffer_size)

        return self


class RouterHttp(_RouterWithForwarders):
    """uWSGI includes an HTTP router/proxy/load-balancer that can forward requests to uWSGI workers.

    The server can be used in two ways:

        * embedded - automatically spawn workers and setup the communication socket
        * standalone - you have to specify the address of a uwsgi socket to connect to

            See `subscribe_to` argument to `.set_basic_params()`

    .. note:: If you want to go massive (virtualhosting and zero-conf scaling) combine the HTTP router
        with the uWSGI Subscription Server.

    """
    alias = 'http'
    plugin = alias

    # todo consider adding
    # http-to-https
    # http-var
    # http-modifier1 / 2
    # http-raw-body - to support chunked input.
    # http-stud-prefix
    # http-server-name-as-http-host
    # http-backend-http
    # http-manage-expect

    def set_basic_params(
            self, *, workers=None, zerg_server=None, fallback_node=None, concurrent_events=None,
            cheap_mode=None, stats_server=None, quiet=None, buffer_size=None,
            keepalive=None, resubscribe_addresses=None):
        """
        :param int workers: Number of worker processes to spawn.

        :param str zerg_server: Attach the router to a zerg server.

        :param str fallback_node: Fallback to the specified node in case of error.

        :param int concurrent_events: Set the maximum number of concurrent events router can manage.

            Default: system dependent.

        :param bool cheap_mode: Enables cheap mode. When the router is in cheap mode,
            it will not respond to requests until a node is available.
            This means that when there are no nodes subscribed, only your local app (if any) will respond.
            When all of the nodes go down, the router will return in cheap mode.

        :param str stats_server: Router stats server address to run at.

        :param bool quiet: Do not report failed connections to instances.

        :param int buffer_size: Set internal buffer size in bytes. Default: page size.

        :param int keepalive: Allows holding the connection open even if the request has a body.

            * http://uwsgi.readthedocs.io/en/latest/HTTP.html#http-keep-alive

            .. note:: See http11 socket type for an alternative.

        :param str|list[str] resubscribe_addresses: Forward subscriptions
            to the specified subscription server.


        """
        super().set_basic_params(**filter_locals(locals(), drop=[
            'keepalive',
            'resubscribe_addresses',
        ]))

        self._set_aliased('keepalive', keepalive)
        self._set_aliased('resubscribe', resubscribe_addresses, multi=True)

        return self

    def set_connections_params(
            self, *, harakiri=None, timeout_socket=None, retry_delay=None, timeout_headers=None, timeout_backend=None):
        """Sets connection-related parameters.

        :param int harakiri: Set gateway harakiri timeout (seconds).

        :param int timeout_socket: Node socket timeout (seconds).
            Used to set the SPDY timeout. This is the maximum amount of inactivity after
            the SPDY connection is closed.

            Default: 60.

        :param int retry_delay: Retry connections to dead static nodes after the specified
            amount of seconds. Default: 30.

        :param int timeout_headers: Defines the timeout (seconds) while waiting for http headers.

            Default: `socket_timeout`.

        :param int timeout_backend: Defines the timeout (seconds) when connecting to backend instances.

            Default: `socket_timeout`.

        """

        super().set_connections_params(
            **filter_locals(locals(), drop=['timeout_headers', 'timeout_backend']))

        self._set_aliased('headers-timeout', timeout_headers)
        self._set_aliased('connect-timeout', timeout_backend)

        return self

    def set_manage_params(
            self, *, chunked_input=None, chunked_output=None, gzip=None, websockets=None, source_method=None,
            rtsp=None, proxy_protocol=None):
        """Allows enabling various automatic management mechanics.

        * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.html#http-router-keepalive-auto-chunking-auto-gzip-and-transparent-websockets

        :param bool chunked_input: Automatically detect chunked input requests and put the session in raw mode.

        :param bool chunked_output: Automatically transform output to chunked encoding
            during HTTP 1.1 keepalive (if needed).

        :param bool gzip: Automatically gzip content if uWSGI-Encoding header is set to gzip,
            but content size (Content-Length/Transfer-Encoding) and Content-Encoding are not specified.

        :param bool websockets: Automatically detect websockets connections and put the session in raw mode.

        :param bool source_method: Automatically put the session in raw mode for `SOURCE` HTTP method.

            * http://uwsgi.readthedocs.io/en/latest/Changelog-2.0.5.html#icecast2-protocol-helpers

        :param bool rtsp: Allow the HTTP router to detect RTSP and chunked requests automatically.

        :param bool proxy_protocol: Allows the HTTP router to manage PROXY1 protocol requests,
            such as those made by Haproxy or Amazon Elastic Load Balancer (ELB).

        """
        self._set_aliased('chunked-input', chunked_input, cast=bool)
        self._set_aliased('auto-chunked', chunked_output, cast=bool)
        self._set_aliased('auto-gzip', gzip, cast=bool)
        self._set_aliased('websockets', websockets, cast=bool)
        self._set_aliased('manage-source', source_method, cast=bool)
        self._set_aliased('manage-rtsp', rtsp, cast=bool)
        self._set_aliased('enable-proxy-protocol', proxy_protocol, cast=bool)

        return self

    def set_owner_params(self, uid=None, gid=None):
        """Drop http router privileges to specified user and group.

        :param str|int uid: Set uid to the specified username or uid.

        :param str|int gid: Set gid to the specified groupname or gid.

        """
        self._set_aliased('uid', uid)
        self._set_aliased('gid', gid)

        return self


class RouterHttps(RouterHttp):
    """uWSGI includes an HTTPS router/proxy/load-balancer that can forward requests to uWSGI workers.

    The server can be used in two ways:

        * embedded - automatically spawn workers and setup the communication socket
        * standalone - you have to specify the address of a uwsgi socket to connect to

            See `subscribe_to` argument to `.set_basic_params()`

    .. note:: If you want to go massive (virtualhosting and zero-conf scaling) combine the HTTP router
        with the uWSGI Subscription Server.

    """
    alias = 'http'  # Shares options with http.
    plugin = alias
    on_command = 'https2'

    def __init__(
            self, on, *, cert, key, ciphers=None, client_ca=None, session_context=None, use_spdy=None,
            export_cert_var=None):
        """Binds https router to run on the given address.

        :param SocketShared|str on: Activates the router on the given address.

        :param str cert: Certificate file.

        :param str key: Private key file.

        :param str ciphers: Ciphers [alias] string.

            Example:
                * DEFAULT
                * HIGH
                * DHE, EDH

            * https://www.openssl.org/docs/man1.1.0/apps/ciphers.html

        :param str client_ca: Client CA file for client-based auth.

            .. note: You can prepend ! (exclamation mark) to make client certificate
                authentication mandatory.

        :param str session_context: Session context identifying string. Can be set to static shared value
            to avoid session rejection.

            Default: a value built from the HTTP server address.

            * http://uwsgi.readthedocs.io/en/latest/SSLScaling.html#setup-2-synchronize-caches-of-different-https-routers

        :param bool use_spdy: Use SPDY.

        :param bool export_cert_var: Export uwsgi variable `HTTPS_CC` containing the raw client certificate.

        """
        on = KeyValue(
            filter_locals(locals(), drop=['session_context']),
            aliases={'on': 'addr', 'use_spdy': 'spdy'},
            bool_keys=['use_spdy'],
        )

        super().__init__(on)

        self._set_aliased('session-context', session_context)


class RouterSsl(_RouterWithForwarders):
    """Works in the same way as the RouterRaw, but will terminate ssl connections.

    Supports SNI for implementing virtual hosting.

    """
    alias = 'sslrouter'
    plugin = alias
    on_command = 'sslrouter2'

    def __init__(self, on, cert, key, forward_to=None, ciphers=None, client_ca=None, session_context=None, use_sni=None):
        """Activates the router on the given address.

        :param SocketShared|str on: Activates the router on the given address.

        :param str cert: Certificate file.

        :param str key: Private key file.

        :param Forwarder|str|list[str] forward_to: Where to forward requests.
            Expects a forwarder instance or one or more node names.

        :param str ciphers: Ciphers [alias] string.

            Example:
                * DEFAULT
                * HIGH
                * DHE, EDH

            * https://www.openssl.org/docs/man1.1.0/apps/ciphers.html

        :param str client_ca: Client CA file for client-based auth.

        :param str session_context: Session context identifying string. Can be set to static shared value
            to avoid session rejection.

            Default: a value built from the HTTP server address.

            * http://uwsgi.readthedocs.io/en/latest/SSLScaling.html#setup-2-synchronize-caches-of-different-https-routers

        :param bool use_sni: Use SNI to route requests.

        """
        on = KeyValue(
            filter_locals(locals(), drop=['session_context', 'use_sni']),
            aliases={'on': 'addr'},
        )

        self._set_aliased('session-context', session_context)
        self._set_aliased('sni', use_sni, cast=bool)

        super().__init__(on, forward_to=forward_to)

    def set_connections_params(self, harakiri=None, timeout_socket=None, retry_delay=None, retry_max=None):
        """Sets connection-related parameters.

        :param int harakiri: Set gateway harakiri timeout (seconds).

        :param int timeout_socket: Node socket timeout (seconds). Default: 60.

        :param int retry_delay: Retry connections to dead static nodes after the specified
            amount of seconds. Default: 30.

        :param int retry_max: Maximum number of retries/fallbacks to other nodes. Default: 3.

        """
        super().set_connections_params(**filter_locals(locals(), drop=['retry_max']))

        self._set_aliased('max-retries', retry_max)

        return self


class RouterFast(_RouterWithForwarders):
    """A proxy/load-balancer/router speaking the uwsgi protocol.

    You can put it between your webserver and real uWSGI instances to have more control
    over the routing of HTTP requests to your application servers.

    """
    alias = 'fastrouter'
    plugin = alias

    def set_basic_params(
            self, *, workers=None, zerg_server=None, fallback_node=None, concurrent_events=None,
            cheap_mode=None, stats_server=None, quiet=None, buffer_size=None,
            fallback_nokey=None, subscription_key=None, emperor_command_socket=None):
        """
        :param int workers: Number of worker processes to spawn.

        :param str zerg_server: Attach the router to a zerg server.

        :param str fallback_node: Fallback to the specified node in case of error.

        :param int concurrent_events: Set the maximum number of concurrent events router can manage.

            Default: system dependent.

        :param bool cheap_mode: Enables cheap mode. When the router is in cheap mode,
            it will not respond to requests until a node is available.
            This means that when there are no nodes subscribed, only your local app (if any) will respond.
            When all of the nodes go down, the router will return in cheap mode.

        :param str stats_server: Router stats server address to run at.

        :param bool quiet: Do not report failed connections to instances.

        :param int buffer_size: Set internal buffer size in bytes. Default: page size.

        :param bool fallback_nokey: Move to fallback node even if a subscription key is not found.

        :param str subscription_key: Skip uwsgi parsing and directly set a key.

        :param str emperor_command_socket: Set the emperor command socket that will receive spawn commands.

            See `.empire.set_emperor_command_params()`.

        """
        super().set_basic_params(**filter_locals(locals(), drop=[
            'fallback_nokey',
            'subscription_key',
            'emperor_command_socket',
        ]))

        self._set_aliased('fallback-on-no-key', fallback_nokey, cast=bool)
        self._set_aliased('force-key', subscription_key)
        self._set_aliased('emperor-socket', emperor_command_socket)

        return self

    def set_resubscription_params(self, addresses=None, bind_to=None):
        """You can specify a dgram address (udp or unix) on which all of the subscriptions
        request will be forwarded to (obviously changing the node address to the router one).

        The system could be useful to build 'federated' setup.

        * http://uwsgi.readthedocs.io/en/latest/Changelog-2.0.1.html#resubscriptions

        :param str|list[str] addresses: Forward subscriptions to the specified subscription server.

        :param str|list[str] bind_to: Bind to the specified address when re-subscribing.

        """
        self._set_aliased('resubscribe', addresses, multi=True)
        self._set_aliased('resubscribe-bind', bind_to)

        return self

    def set_connections_params(self, harakiri=None, timeout_socket=None, retry_delay=None, retry_max=None, defer=None):
        """Sets connection-related parameters.

        :param int harakiri:  Set gateway harakiri timeout (seconds).

        :param int timeout_socket: Node socket timeout (seconds). Default: 60.

        :param int retry_delay: Retry connections to dead static nodes after the specified
            amount of seconds. Default: 30.

        :param int retry_max: Maximum number of retries/fallbacks to other nodes. Default: 3

        :param int defer: Defer connection delay, seconds. Default: 5.

        """
        super().set_connections_params(**filter_locals(locals(), drop=['retry_max', 'defer']))

        self._set_aliased('max-retries', retry_max)
        self._set_aliased('defer-connect-timeout', defer)

        return self

    def set_postbuffering_params(self, size=None, store_dir=None):
        """Sets buffering params.

        Web-proxies like nginx are "buffered", so they wait til the whole request (and its body)
        has been read, and then it sends it to the backends.

        :param int size: The size (in bytes) of the request body after which the body will
            be stored to disk (as a temporary file) instead of memory.

        :param str store_dir: Put buffered files to the specified directory. Default: TMPDIR, /tmp/

        """
        self._set_aliased('post-buffering', size)
        self._set_aliased('post-buffering-dir', store_dir)

        return self

    def set_owner_params(self, uid=None, gid=None):
        """Drop http router privileges to specified user and group.

        :param str|int uid: Set uid to the specified username or uid.

        :param str|int gid: Set gid to the specified groupname or gid.

        """
        self._set_aliased('uid', uid)
        self._set_aliased('gid', gid)

        return self


class RouterRaw(_RouterWithForwarders):
    """A pure-TCP load balancer.

    Can be used to load balance between the various HTTPS routers.

    """
    alias = 'rawrouter'
    plugin = alias

    def set_connections_params(
            self, harakiri=None, timeout_socket=None, retry_delay=None, retry_max=None, use_xclient=None):
        """Sets connection-related parameters.

        :param int harakiri:  Set gateway harakiri timeout (seconds).

        :param int timeout_socket: Node socket timeout (seconds). Default: 60.

        :param int retry_delay: Retry connections to dead static nodes after the specified
            amount of seconds. Default: 30.

        :param int retry_max: Maximum number of retries/fallbacks to other nodes. Default: 3.

        :param bool use_xclient: Use the xclient protocol to pass the client address.

        """
        super().set_connections_params(**filter_locals(locals(), drop=['retry_max', 'use_xclient']))

        self._set_aliased('max-retries', retry_max)
        self._set_aliased('xclient', use_xclient)

        return self


class RouterForkPty(_RouterCommon):
    """Allows allocation of pseudoterminals in jails.

    Dealing with containers is now a common deployment pattern.
    One of the most annoying tasks when dealing with jails/namespaces is 'attaching' to already running instances.
    The forkpty router aims at simplifyng the process giving a pseudoterminal server to your uWSGI instances.
    A client connect to the socket exposed by the forkpty router and get a new pseudoterminal
    connected to a process (generally a shell, but can be whatever you want).

    .. note:: To be used in cooperation with `pty` plugin.

    """
    alias = 'forkptyrouter'
    plugin = alias

    def __init__(self, on=None, undeferred=False):
        """Binds router to run on the given address.

        :param SocketShared|str on: Activates the router on the given address.

        :param bool undeferred: Run router in undeferred mode.

        """
        router_name = self.alias

        if undeferred:
            router_name = 'forkptyurouter'

        self.on_command = router_name

        super().__init__(on)

    def set_basic_params(
            self, *, workers=None, zerg_server=None, fallback_node=None, concurrent_events=None,
            cheap_mode=None, stats_server=None, run_command=None):
        """

        :param int workers: Number of worker processes to spawn.

        :param str zerg_server: Attach the router to a zerg server.

        :param str fallback_node: Fallback to the specified node in case of error.

        :param int concurrent_events: Set the maximum number of concurrent events router can manage.

            Default: system dependent.

        :param bool cheap_mode: Enables cheap mode. When the router is in cheap mode,
            it will not respond to requests until a node is available.
            This means that when there are no nodes subscribed, only your local app (if any) will respond.
            When all of the nodes go down, the router will return in cheap mode.

        :param str stats_server: Router stats server address to run at.

        :param str run_command: Run the specified command
            on every connection. Default: /bin/sh.

        """
        super().set_basic_params(**filter_locals(locals(), drop=['run_command']))

        self._set_aliased('command', run_command)

        return self

    def set_connections_params(self, *, harakiri=None, timeout_socket=None):
        """Sets connection-related parameters.

        :param int harakiri: Set gateway harakiri timeout (seconds).

        :param int timeout_socket: Node socket timeout (seconds). Default: 60.

        """
        self._set_aliased('harakiri', harakiri)
        self._set_aliased('timeout', timeout_socket)

        return self

    def set_window_params(self, cols=None, rows=None):
        """Sets pty window params.

        :param int cols:
        :param int rows:

        """
        self._set_aliased('cols', cols)
        self._set_aliased('rows', rows)

        return self


class RouterTunTap(RouterBase):
    """The tuntap router is a non-blocking highly optimized ip router
    translating from tuntap device to socket streams.

    Allows full user-space networking in jails.

    It is meant as a replacement for the currently available networking namespaces approaches.
    Compared to `veth` or `macvlan` it is really simple and allows total control over the routing subsystem
    (in addition to a simple customizable firewalling engine).

    Generally you spawn the tuntap router in the Emperor instance.
    Vassals will run in new namespaces in which they create a tuntap device attached to the tuntap router.
    UNIX sockets are the only way to connect to the tuntap router after jailing.

    Vassals should connect to tuntap device.

    """
    alias = 'tuntap'
    plugin = alias

    def __init__(self, on=None, *, device=None, stats_server=None, gateway=None):
        """Passing params will create a router device.

        :param str on: Socket file.

        :param str device: Device name.

        :param str stats_server: Router stats server address to run at.

        :param str gateway: Gateway address.

        """
        super().__init__()

        if on is not None:
            value = [device or 'uwsgidev', on]

            if stats_server:
                value.append(stats_server)

                if gateway:
                    value.append(gateway)

            self._set_aliased('router', ' '.join(value), multi=True)

    def set_basic_params(self, *, use_credentials=None, stats_server=None):
        """
        :param str use_credentials: Enable check of SCM_CREDENTIALS for tuntap client/server.

        :param str stats_server: Router stats server address to run at.

        """
        self._set_aliased('use-credentials', use_credentials)
        self._set_aliased('router-stats', stats_server)

        return self

    def register_route(self, src, dst, *, gateway):
        """Adds a routing rule to the tuntap router.

        :param str src: Source/mask.

        :param str dst: Destination/mask.

        :param str gateway: Gateway address.

        """
        self._set_aliased('router-route', ' '.join((src, dst, gateway)), multi=True)

        return self

    def device_connect(self, socket, *, device_name):
        """Add a tuntap device to the instance.

        To be used in a vassal.

        :param str socket: Router socket.

            Example: `/run/tuntap_router.socket`.

        :param str device_name: Device.

            Example: `uwsgi0`.

        """
        self._set_aliased('device', f'{device_name} {socket}')

        return self

    def device_add_rule(self, *, direction, action, src, dst, target=None):
        """Adds a tuntap device rule.

        To be used in a vassal.

        :param str direction: Direction:

            * in
            * out.

        :param str action: Action:

            * allow
            * deny
            * route
            * gateway.

        :param str src: Source/mask.

        :param str dst: Destination/mask.

        :param str target: Depends on action.

            * Route / Gateway: Accept addr:port

        """
        value = [direction, src, dst, action]

        if target:
            value.append(target)

        self._set_aliased('device-rule', ' '.join(value), multi=True)

        return self

    def add_firewall_rule(self, *, direction, action, src=None, dst=None):
        """Adds a firewall rule to the router.

        The TunTap router includes a very simple firewall for governing vassal's traffic.
        The first matching rule stops the chain, if no rule applies, the policy is "allow".

        :param str direction: Direction:

            * in
            * out

        :param str action: Action:

            * allow
            * deny

        :param str src: Source/mask.

        :param str dst: Destination/mask

        """
        value = [action]

        if src:
            value.extend((src, dst))

        self._set_aliased(f'router-firewall-{direction.lower()}', ' '.join(value), multi=True)

        return self



from django import db
from django.apps import apps
from django.utils.module_loading import autodiscover_modules

from uwsgiconf import uwsgi
from uwsgiconf.exceptions import RuntimeConfigurationError
from uwsgiconf.settings import FORCE_STUB
from .settings import MODULE_INIT, MODULE_INIT_DEFAULT


def check_for_stub():
    """Check for uWSGI stub. Disallow it to prevent
    stub module caching when embedded mode with pyuwsgi is used."""

    if not uwsgi.is_stub:
        # Native uwsgi module.
        return

    if FORCE_STUB:
        # Stub is used deliberately (e.g. in a test suite).
        return

    msg = (
        'Something from uwsgiconf.uwsgi has been imported before uWSGI start. '
        'Please move uWSGI related stuff including such imports '
        f'into {MODULE_INIT}.py modules of your apps.'
    )

    raise RuntimeConfigurationError(msg)


check_for_stub()


from uwsgiconf.runtime.platform import uwsgi


@uwsgi.postfork_hooks.add()
def db_close_connections():
    """Close db connections after fork()."""
    db.connections.close_all()


if apps.apps_ready:

    if MODULE_INIT != MODULE_INIT_DEFAULT:
        # Import uWSGI init modules from applications.
        autodiscover_modules(MODULE_INIT)



import logging
import os
import sys
from collections import namedtuple
from contextlib import contextmanager
from importlib import import_module
from io import StringIO
from types import ModuleType
from typing import Optional, Any, List, Tuple, Dict

from .exceptions import UwsgiconfException
from .settings import CONFIGS_MODULE_ATTR
from .typehints import Strlist

if False:  # pragma: nocover
    from .config import Configuration  # noqa


EmbeddedPlugins = namedtuple('EmbeddedPlugins', ['generic', 'request'])


def get_logger(name: str):
    # Here to mitigate module name clashing.
    return logging.getLogger(name)


def encode(value: str) -> bytes:
    """Encodes str into bytes."""
    return value.encode('utf-8')


def decode(value: Optional[bytes]) -> Optional[str]:
    """Decodes bytes into str."""
    if value is None:
        return value
    return value.decode('utf-8')


def decode_deep(value: Any) -> Any:
    """Decodes object deep if required.

    :param value:

    """
    if isinstance(value, dict):
        out = {}
        for key, val in value.items():
            out[decode_deep(key)] = decode_deep(val)

    elif isinstance(value, tuple):
        out = tuple(decode_deep(item) for item in value)

    elif isinstance(value, list):
        out = [decode_deep(item) for item in value]

    elif isinstance(value, bytes):
        out = decode(value)

    else:
        out = value

    return out


@contextmanager
def output_capturing():
    """Temporarily captures/redirects stdout."""
    out = sys.stdout

    sys.stdout = StringIO()

    try:
        yield

    finally:
        sys.stdout = out


class ConfModule:
    """Represents a uwsgiconf configuration module.

    Allows reading configurations from .py files.

    """
    default_name: str = 'uwsgicfg.py'

    def __init__(self, fpath: str):
        """Module filepath.

        :param fpath:

        """
        fpath = os.path.abspath(fpath)
        self.fpath = fpath
        self._confs = None

    def spawn_uwsgi(self, *, only: str = None) -> List[Tuple[str, int]]:
        """Spawns uWSGI process(es) which will use configuration(s) from the module.

        Returns list of tuples:
            (configuration_alias, uwsgi_process_id)

        If only one configuration found current process (uwsgiconf) is replaced with a new one (uWSGI),
        otherwise a number of new detached processes is spawned.

        :param only: Configuration alias to run from the module.
            If not set uWSGI will be spawned for every configuration found in the module.

        """
        spawned = []
        configs = self.configurations

        if len(configs) == 1:

            alias = configs[0].alias
            UwsgiRunner().spawn(configs[0], replace=True, filepath=self.fpath)
            spawned.append((alias, os.getpid()))

        else:
            for config in configs:
                alias = config.alias

                if only is None or alias == only:
                    pid = UwsgiRunner().spawn(config, filepath=self.fpath)
                    spawned.append((alias, pid))

        return spawned

    @property
    def configurations(self) -> List['Configuration']:
        """Configurations from uwsgiconf module."""

        if self._confs is not None:
            return self._confs

        with output_capturing():
            module = self.load(self.fpath)
            confs = getattr(module, CONFIGS_MODULE_ATTR)
            confs = listify(confs)

        self._confs = confs

        return confs

    @classmethod
    def load(cls, fpath: str) -> ModuleType:
        """Loads a module and returns its object.

        :param fpath:

        """
        module_name = os.path.splitext(os.path.basename(fpath))[0]

        sys.path.insert(0, os.path.dirname(fpath))
        try:
            module = import_module(module_name)

        finally:
            sys.path = sys.path[1:]

        return module


def listify(src: Any) -> List:
    """Make a list with source object if not already a list.

    :param src:

    """
    if not isinstance(src, list):
        src = [src]

    return src


def filter_locals(
        locals_dict: Dict[str, Any],
        *,
        drop: List[str] = None,
        include: List[str] = None
) -> Dict[str, Any]:
    """Filters a dictionary produced by locals().

    :param locals_dict:

    :param drop: Keys to drop from dict.

    :param include: Keys to include into dict.

    """
    drop = drop or []
    drop.extend([
        'self',
        '__class__',  # py3
    ])

    include = include or locals_dict.keys()

    relevant_keys = [key for key in include if key not in drop]

    locals_dict = {k: v for k, v in locals_dict.items() if k in relevant_keys}

    return locals_dict


class KeyValue:
    """Allows lazy flattening the given dictionary into a key-value string."""

    def __init__(
            self,
            locals_dict: Dict[str, Any],
            *,
            keys: List[str] = None,
            aliases: Dict[str, str] = None,
            bool_keys: List[str] = None,
            list_keys: List[str] = None,
            items_separator: str = ','
    ):
        """

        :param locals_dict: Dictionary produced by locals().

        :param keys: Relevant keys from dictionary.
            If not defined - all keys are relevant.
            If defined keys will flattened into string using given order.

        :param aliases: Mapping key names from locals_dict into names
            they should be replaced with.

        :param bool_keys: Keys to consider their values bool.

        :param list_keys: Keys expecting lists.

        :param items_separator: String to use as items (chunks) separator.

        """
        self.locals_dict = dict(locals_dict)
        self.keys = keys or sorted(filter_locals(locals_dict).keys())
        self.aliases = aliases or {}
        self.bool_keys = bool_keys or []
        self.list_keys = list_keys or []
        self.items_separator = items_separator

    def __str__(self):
        value_chunks = []

        for key in self.keys:
            val = self.locals_dict[key]

            if val is not None:

                if key in self.bool_keys:
                    val = 1

                elif key in self.list_keys:

                    val = ';'.join(listify(val))

                value_chunks.append(f'{self.aliases.get(key, key)}={val}')

        return self.items_separator.join(value_chunks).strip()


def get_output(cmd: str, *, args: Strlist) -> str:
    """Runs a command and returns its output (stdout + stderr).

    :param cmd:
    :param args:

    """
    from subprocess import Popen, STDOUT, PIPE

    command = [cmd]
    command.extend(listify(args))

    process = Popen(command, stdout=PIPE, stderr=STDOUT)
    out, _ = process.communicate()

    return out.decode('utf-8')


class Finder:
    """Finds various entities."""

    @classmethod
    def uwsgiconf(cls) -> str:
        """Finds uwsgiconf executable location."""
        return get_output('which', args=['uwsgiconf']).strip()

    @classmethod
    def python(cls) -> str:
        """Finds Python executable location."""
        return sys.executable


class Fifo:
    """uWSGI Master FIFO interface."""

    def __init__(self, fifo_filepath: str):
        """
        :param fifo_filepath: Path to uWSGI Master FIFO file.

        """
        self.fifo = fifo_filepath

    def cmd_log(self, *, reopen: bool = False, rotate: bool = False):
        """Allows managing of uWSGI log related stuff

        :param reopen: Reopen log file. Could be required after third party rotation.
        :param rotate: Trigger built-in log rotation.

        """
        cmd = b''

        if reopen:
            cmd += b'l'

        if rotate:
            cmd += b'L'

        return self.send_command(cmd)

    def cmd_stats(self):
        """Dump uWSGI configuration and current stats into the log."""
        return self.send_command(b's')

    def cmd_stop(self, *, force: bool = False):
        """Shutdown uWSGI instance.

        :param force: Use forced (brutal) shutdown instead of a graceful one.

        """
        return self.send_command(b'Q' if force else b'q')

    def cmd_reload(self, *, force: bool = False, workers_only: bool = False, workers_chain: bool = False):
        """Reloads uWSGI master process, workers.

        :param force: Use forced (brutal) reload instead of a graceful one.
        :param workers_only: Reload only workers.
        :param workers_chain: Run chained workers reload (one after another,
            instead of destroying all of them in bulk).

        """
        if workers_chain:
            return self.send_command(b'c')

        if workers_only:
            return self.send_command(b'R' if force else b'r')

        return self.send_command(b'R' if force else b'r')

    def send_command(self, cmd: bytes):
        """Sends a generic command into FIFO.

        :param cmd: Command chars to send into FIFO.

        """
        if not cmd:
            return

        with open(self.fifo, 'wb') as f:
            f.write(cmd)


class UwsgiRunner:
    """Exposes methods to run uWSGI."""

    def __init__(self, binary_path: str = None):
        self.binary_uwsgi = binary_path or 'uwsgi'
        self.binary_python = self.prepare_env()

    def get_output(self, command_args: Strlist) -> str:
        """Runs a command and returns its output (stdout + stderr).

        :param command_args:

        """
        return get_output(self.binary_uwsgi, args=command_args)

    def get_plugins(self) -> EmbeddedPlugins:
        """Returns ``EmbeddedPlugins`` object with."""
        out = self.get_output('--plugin-list')
        return parse_command_plugins_output(out)

    @classmethod
    def get_env_path(cls) -> str:
        """Returns PATH environment variable updated
        to run uwsgiconf in (e.g. for virtualenv).

        """
        return os.path.dirname(Finder.python()) + os.pathsep + os.environ['PATH']

    @classmethod
    def prepare_env(cls) -> str:
        """Prepares current environment and returns Python binary name.

        This adds some virtualenv friendliness so that we try use uwsgi from it.

        """
        os.environ['PATH'] = cls.get_env_path()
        return os.path.basename(Finder.python())

    def spawn(
            self,
            config: 'Configuration',
            *,
            replace: bool = False,
            filepath: str = None,
            embedded: bool = False
    ):
        """Spawns uWSGI using the given configuration module.

        .. note::
            uWSGI loader schemas:
                * From a symbol -- sym://uwsgi_funny_function
                * From binary appended data -- data://0
                * From http -- http://example.com/hello
                * From a file descriptor -- fd://3
                * From a process stdout -- exec://foo.pl
                * From a function call returning a char * -- call://uwsgi_func

            Loading in config:
                foo = @(sym://uwsgi_funny_function)

        :param config: Configuration object to spawn uWSGI with.

        :param filepath: Override configuration file path.

        :param replace: Whether a new process should replace current one.

        :param embedded: Flag. Do not create a configuration file even if required,
            translate all config parameters into command line arguments and
            pass it to ``pyuwsgi``.

        """
        args = ['uwsgi']

        if embedded:
            import pyuwsgi  # noqa
            args.extend(config.format(formatter='args'))
            pyuwsgi.run(args[1:])

        else:
            args.append('--ini')

            filepath = filepath or config.tofile()

            if os.path.splitext(os.path.basename(filepath))[1] == '.ini':
                args.append(filepath)

            else:
                # Consider it to be a python script (uwsgicfg.py).
                # Pass --conf as an argument to have a chance to use
                # touch reloading form .py configuration file change.
                args.append(f'exec://{self.binary_python} {filepath} --conf {config.alias}')

        if replace:

            try:
                return os.execvp('uwsgi', args)

            except FileNotFoundError:

                raise UwsgiconfException(
                    'uWSGI executable not found. '
                    'Please make sure it is installed and available.')

        return os.spawnvp(os.P_NOWAIT, 'uwsgi', args)


def parse_command_plugins_output(out: str) -> EmbeddedPlugins:
    """Parses ``plugin-list`` command output from uWSGI
    and returns object containing lists of embedded plugin names.

    :param out:

    """
    out = out.split('--- end of plugins list ---')[0]
    out = out.partition('plugins ***')[2]
    out = out.splitlines()

    current_slot = 0

    plugins = EmbeddedPlugins([], [])

    for line in out:
        line = line.strip()

        if not line:
            continue

        if line.startswith('***'):
            current_slot += 1
            continue

        if current_slot is not None:
            plugins[current_slot].append(line)

    plugins = plugins._replace(request=[plugin.partition(' ')[2] for plugin in plugins.request])

    return plugins


def get_uwsgi_stub_attrs_diff() -> Tuple[List[str], List[str]]:
    """Returns attributes difference two elements tuple between
    real uwsgi module and its stub.

    Might be of use while describing in stub new uwsgi functions.

    Returns (uwsgi_only_attrs, stub_only_attrs)

    """
    try:
        import uwsgi

    except ImportError:
        from uwsgiconf.exceptions import UwsgiconfException

        raise UwsgiconfException(
            '`uwsgi` module is unavailable. Calling `get_attrs_diff` in such environment makes no sense.')

    from . import uwsgi_stub

    def get_attrs(src):
        return set(attr for attr in dir(src) if not attr.startswith('_'))

    attrs_uwsgi = get_attrs(uwsgi)
    attrs_stub = get_attrs(uwsgi_stub)

    from_uwsgi = sorted(attrs_uwsgi.difference(attrs_stub))
    from_stub = sorted(attrs_stub.difference(attrs_uwsgi))

    return from_uwsgi, from_stub



import pytest

from uwsgiconf.runtime.scheduling import *


def test_timers():

    @register_timer(20)
    def fire1():
        pass

    @register_timer_rb(3)
    def fire2():
        pass


@pytest.mark.xfail(reason='Flacky fail due to a date change. Need to change test someday')
def test_cron():

    @register_cron(hour=-3)
    def fire1():
        pass

    with pytest.raises(RuntimeConfigurationError):
        register_cron(hour='-%s/2')

    # todo this test is flacky due on variable date. it needs the freezegun.
    now = datetime.now()

    @register_cron(hour='%s-%s/2' % (now.hour, now.hour+3), weekday='0-6')
    def runnable1():
        return 100

    @register_cron(hour='%s-%s' % (now.hour, now.hour+3))
    def runnable2():
        return 200

    @register_cron(hour='%s-%s' % (now.hour-3, now.hour-1))
    def not_runnable():
        return 200

    assert runnable1() == 100
    assert runnable2() == 200
    assert not_runnable() is None



from typing import Callable, Sequence, Tuple, List, Optional

from .. import uwsgi
from ..utils import encode, decode, get_logger, decode_deep

_LOG = get_logger(__name__)


def register_rpc(name: str = None) -> Callable:
    """Decorator. Allows registering a function for RPC.

    * http://uwsgi.readthedocs.io/en/latest/RPC.html

    .. code-block:: python

        @register_rpc()
        def expose_me(arg1, arg2=15):
            print(f'RPC called {arg1}')
            return b'some'

        make_rpc_call('expose_me', ['value1'])

    .. warning:: Function expected to accept bytes args.
        Also expected to return bytes or ``None``.

    :param name: RPC function name to associate
        with decorated function.

    """
    def wrapper(func: Callable):
        func_name = func.__name__
        rpc_name = name or func_name

        uwsgi.register_rpc(rpc_name, func)

        _LOG.debug(f"Registering '{func_name}' for RPC under '{rpc_name}' alias ...")

        return func

    return wrapper


def make_rpc_call(func_name: str, *, args: Sequence[str] = None, remote: str = None) -> Optional[str]:
    """Performs an RPC function call (local or remote) with the given arguments.

    :param func_name: RPC function name to call.

    :param Iterable args: Function arguments.

        .. warning:: Strings are expected.

    :param remote:

    :raises ValueError: If unable to call RPC function.

    """
    args = args or []
    args = [encode(str(arg)) for arg in args]

    func_name = encode(func_name)

    if remote:
        result = uwsgi.rpc(encode(remote), func_name, *args)

    else:
        result = uwsgi.call(func_name, *args)

    return decode(result)


def get_rpc_list() -> List[str]:
    """Returns registered RPC functions names."""
    return decode_deep(uwsgi.rpc_list())




class UwsgiconfException(Exception):
    """Base for exceptions."""


class ConfigurationError(UwsgiconfException):
    """Configuration related error."""


class RuntimeConfigurationError(ConfigurationError):
    """Runtime configuration related error."""



from uwsgiconf.config import Section
from uwsgiconf.presets.empire import Broodlord


def test_broodlord(assert_lines):

    emperor, zerg = Broodlord(

        zerg_socket='/tmp/broodlord.sock',
        zerg_count=40,
        zerg_die_on_idle=30,
        vassals_home='/etc/vassals',
        vassal_queue_items_sos=10,

        section_emperor=(Section().
            master_process(enable=True).
            workers(count=1).
            logging(no_requests=True).
            python.set_wsgi_params(module='werkzeug.testapp:test_app')
        )

    ).configure()

    emperor.networking.register_socket(Section.networking.sockets.default(':3031'))

    assert_lines([
        'master = true',
        'workers = 1',
        'disable-logging = true',
        'wsgi = werkzeug.testapp:test_app',
        'zerg-server = /tmp/broodlord.sock',
        'emperor = /etc/vassals',
        'emperor-broodlord = 40',
        'vassal-sos-backlog = 10',
        'socket = :3031',
    ], emperor)

    assert_lines([
        'master = true',
        'workers = 1',
        'disable-logging = true',
        'wsgi = werkzeug.testapp:test_app',
        'idle = 30',
        'die-on-idle = true',
    ], zerg)



from os import environ

from uwsgiconf.presets.nice import Section, PythonSection


def test_nice_section(assert_lines):

    assert_lines([
        'env = LANG=en_US.UTF-8',
        'workers = %k',
        'die-on-term = true',
        'vacuum = true',
        'threads = 4',

    ], Section(threads=4))

    assert_lines([
        'logto',
    ], Section(), assert_in=False)

    assert_lines([
        'enable-threads = true',
        'uid = www-data',
        'gid = www-data',
        'logto = /a/b.log',

    ], Section(threads=True, log_into='/a/b.log').configure_owner())

    assert_lines([
        'workers = 13',
        'touch-reload', 'test_nice.py',

    ], Section(workers=13, touch_reload=__file__))

    assert_lines([
        'disable-write-exception = true',
        'ignore-write-errors = true',
        'ignore-sigpipe = true',
        'log-master = true',
        'threaded-logger = true',

    ], Section(log_dedicated=True, ignore_write_errors=True))

    assert '%(headers) headers in %(hsize) bytes' in Section().get_log_format_default()


def test_get_bundled_static_path(assert_lines):

    path = Section.get_bundled_static_path('503.html')
    assert path.endswith('uwsgiconf/contrib/django/uwsgify/static/uwsgify/503.html')


def test_configure_https_redirect(assert_lines):

    section = Section()
    section.configure_https_redirect()
    assert_lines(
        'route-if-not = eq:${HTTPS};on redirect-301:https://${HTTP_HOST}${REQUEST_URI}',
        section
    )


def test_configure_maintenance_mode(assert_lines, tmpdir):

    section = Section()
    section.configure_maintenance_mode('/watch/that/file', '/serve/this/file')
    section.configure_maintenance_mode('/watch/that/file/also', 'http://pythonz.net')

    assert_lines([
        'route-if = exists:/watch/that/file static:/serve/this/file',
        'route-if = exists:/watch/that/file/also redirect-302:http://pythonz.net',

    ], section)

    afile = tmpdir.join('maintenance_file')

    section = Section()
    section.configure_maintenance_mode(f'{afile}', 'app')
    assert_lines([
        f'env = UWSGICONF_MAINTENANCE={afile}',
        f'touch-reload = {afile}',
    ], section)

    assert_lines([
        'wsgi = uwsgiconf.maintenance:app_maintenance',
    ], section, assert_in=False)

    # Create file
    afile.write('')
    section = Section()
    section.configure_maintenance_mode(f'{afile}', 'app')
    assert_lines([
        f'env = UWSGICONF_MAINTENANCE={afile}',
        f'touch-reload = {afile}',
        'env = UWSGICONF_MAINTENANCE_INPLACE=1',
        'wsgi = uwsgiconf.maintenance:app_maintenance',
    ], section)

    assert environ['UWSGICONF_MAINTENANCE'] == f'{afile}'
    assert environ['UWSGICONF_MAINTENANCE_INPLACE'] == '1'

    section.configure_maintenance_mode(f'{afile}', 'app::mypack.here.there:myfunc')
    assert_lines([
        'wsgi = mypack.here.there:myfunc',
    ], section)


def test_configure_logging_json(assert_lines):

    section = Section()
    section.configure_logging_json()

    assert_lines([
        'logger-req = stdio:',
        'log-format = %(method) %(uri) -> %(status)',
        'log-req-encoder = json {"dt": "${strftime:%%Y-%%m-%%dT%%H:%%M:%%S%%z}", "src": "uwsgi.req"',
        'log-req-encoder = nl',
        '"src": "uwsgi.out"',

    ], section)


def test_configure_certbot_https(assert_lines, monkeypatch):

    monkeypatch.setattr('pathlib.Path.exists', lambda self: True)

    section = Section()
    section.configure_certbot_https('mydomain.org', '/var/www/', address=':4443')

    assert_lines([
        'static-map2 = /.well-known/=/var/www/',
        'https-socket = :4443,/etc/letsencrypt/live/mydomain.org/fullchain.pem,'
        '/etc/letsencrypt/live/mydomain.org/privkey.pem',
    ], section)

    section = Section.bootstrap(['http://:80'])
    section.configure_certbot_https('mydomain.org', '/var/www/', http_redirect=True)

    assert_lines([
        'shared-socket = :80',
        'shared-socket = :443',
        'http-socket = =0',
        'https-socket = =1,/etc/letsencrypt/live/mydomain.org/fullchain.pem,'
        '/etc/letsencrypt/live/mydomain.org/privkey.pem',
        'route-if-not = eq:${HTTPS};on redirect-301:https://${HTTP_HOST}${REQUEST_URI}',
    ], section)


def test_nice_python(assert_lines):

    assert_lines([
        'plugin = python',
        'pyhome = /home/idle/venv/\npythonpath = /home/idle/apps/',
        'wsgi = somepackage.module',
        'need-app = true',

    ], PythonSection(
        params_python=dict(
            # We'll run our app using virtualenv.
            python_home='/home/idle/venv/',
            search_path='/home/idle/apps/',
        ),
        wsgi_module='somepackage.module',
        embedded_plugins=None
    ))

    # Embedded plugins = True
    assert_lines('plugin = python', PythonSection(wsgi_module='somepackage.module'), assert_in=False)



from pathlib import Path
from typing import Union

from ..base import ParametrizedValue
from ..utils import listify


class HookAction(ParametrizedValue):

    pass


class ActionMount(HookAction):
    """Mount or unmount filesystems.

    Examples:
        * Mount: proc none /proc
        * Unmount: /proc

    """

    name = 'mount'

    def __init__(self, mountpoint, *, fs=None, src=None, flags=None):
        """

        :param str mountpoint:

        :param str fs: Filesystem. Presence indicates mounting.

        :param str src: Presence indicates mounting.

        :param str|list flags: Flags available for the operating system.
            As an example on Linux you will options like: bind, recursive, readonly, rec, detach etc.

        """
        if flags is not None:
            flags = listify(flags)
            flags = ','.join(flags)

        if fs:
            args = [fs, src, mountpoint, flags]

        else:
            args = [mountpoint, flags]
            self.name = 'umount'

        super().__init__(*args)


class ActionExecute(HookAction):
    """Run the shell command.

    Command run under ``/bin/sh``.
    If for some reason you do not want to use ``/bin/sh``,
    use ``binsh`` option,

    Examples:
        * cat /proc/self/mounts

    """

    name = 'exec'

    # todo consider adding safeexec
    def __init__(self, command):
        super().__init__(command)


class ActionCall(HookAction):
    """Call functions in the current process address space."""

    name = 'call'

    def __init__(self, target, *, honour_exit_status=False, arg_int=False):
        """
        :param str target: Symbol and args.

        :param bool honour_exit_status: Expect an int return.
            Anything != 0 means failure.

        :param bool arg_int: Parse the argument as an int.

        """
        name = self.name

        if arg_int:
            name += 'int'

        if honour_exit_status:
            name += 'ret'

        self.name = name

        super().__init__(target)


class ActionDirChange(HookAction):
    """Changes a directory.

    Convenience action, same as ``call:chdir <directory>``.

    """
    name = 'cd'

    def __init__(self, target_dir):
        super().__init__(target_dir)


class ActionDirCreate(HookAction):
    """Creates a directory with 0777."""

    name = 'mkdir'

    def __init__(self, target_dir):
        super().__init__(target_dir)


class ActionFileCreate(HookAction):
    """Creates a directory with 0666."""

    name = 'create'

    def __init__(self, fpath: Union[str, Path]):
        super().__init__(fpath)


class ActionExit(HookAction):
    """Exits.

    Convenience action, same as ``callint:exit [num]``.

    """
    name = 'exit'

    def __init__(self, status_code=None):
        super().__init__(status_code)


class ActionPrintout(HookAction):
    """Prints.

    Convenience action, same as calling the ``uwsgi_log`` symbol.

    """
    name = 'print'

    def __init__(self, text=None):
        super().__init__(text)


class ActionSetHostName(HookAction):
    """Sets a host name."""

    name = 'hostname'

    def __init__(self, name):
        super().__init__(name)


class ActionAlarm(HookAction):
    """Issues an alarm. See ``.alarms`` options group."""

    name = 'alarm'

    def __init__(self, alarm, message):
        super().__init__(alarm, message)


class ActionFileWrite(HookAction):
    """Writes a string to the specified file.

    If file doesn't exist it will be created.

    .. note:: Since 1.9.21

    """
    name = 'write'

    def __init__(self, target, text, *, append=False, newline=False):
        """

        :param str target: File to write to.

        :param str text: Text to write into file.

        :param bool append: Append text instead of rewrite.

        :param bool newline: Add a newline at the end.

        """
        if append:
            self.name = 'append'

        if newline:
            self.name += 'n'

        super().__init__(target, text)


class ActionFifoWrite(HookAction):
    """Writes a string to the specified FIFO (see ``fifo_file`` from ``master_process`` params)."""

    name = 'writefifo'

    def __init__(self, target, text, *, wait=False):
        """
        :param bool wait: Wait until FIFO is available.

        """
        if wait:
            self.name = 'spinningfifo'

        super().__init__(target, text)


class ActionUnlink(HookAction):
    """Unlink the specified file.

    .. note:: Since 1.9.21

    """
    name = 'unlink'

    def __init__(self, target):
        super().__init__(target)



from os.path import dirname, basename, abspath
from textwrap import dedent
from typing import List, Dict, Callable, Union

from .config import Configuration, Section
from .utils import Finder, UwsgiRunner


TYPE_UPSTART = 'upstart'
TYPE_SYSTEMD = 'systemd'

TYPES: List[str] = [
    TYPE_UPSTART,
    TYPE_SYSTEMD,
]


def get_tpl_systemd(conf: 'Section') -> str:
    """

    Some Systemd hints:

        * uwsgiconf sysinit > my.service
        * sudo cp my.service /etc/systemd/system/

        * sudo sh -c "systemctl daemon-reload; systemctl start my.service"

        * journalctl -fu my.service

    :param conf: Section object.

    """
    tpl = '''
        # Place into:   /etc/systemd/system/{project}.service
        # Setup:        sudo systemctl enable --now $PWD/{project}.service
        # Start:        sudo systemctl start {project}.service
        # Stop:         sudo systemctl stop {project}.service
        # Restart:      sudo systemctl restart {project}.service
        # Status:       systemctl status {project}.service
        # Journal:      journalctl -fu {project}.service

        [Unit]
        Description={project} uWSGI Service
        Wants=network-online.target
        After=network-online.target

        [Service]
        Environment="PATH=%(path)s"
        ExecStartPre=-/usr/bin/install -d -m 0755 -o %(user)s -g %(group)s %(runtime_dir)s
        ExecStart={command}
        Restart=on-failure
        KillSignal=SIGTERM
        Type=notify
        StandardError=syslog
        NotifyAccess=all
        # Bind to priviledged ports.
        AmbientCapabilities=CAP_NET_BIND_SERVICE

        [Install]
        WantedBy=multi-user.target
    '''
    # We do not use 'RuntimeDirectory' systemd directive since we need to chown.

    uid, gid = conf.main_process.get_owner()

    tpl = tpl % {
        'runtime_dir': conf.replace_placeholders('{project_runtime_dir}'),
        'path': UwsgiRunner.get_env_path(),
        'user':  uid,
        'group': gid,
    }

    return tpl


def get_tpl_upstart(conf: 'Section') -> str:
    """

    :param conf: Section object.

    """
    tpl = '''
        # Place into: /etc/init/{project}.conf
        # Verify:     initctl check-config {project}
        # Start:      initctl start {project}
        # Stop:       initctl stop {project}
        # Restart:    initctl restart {project}
        
        description "{project} uWSGI Service"
        start on runlevel [2345]
        stop on runlevel [06]
        
        respawn
        
        env PATH=%(path)s
        pre-start exec -/usr/bin/install -d -m 0755 -o %(user)s -g %(group)s %(runtime_dir)s
        
        exec {command}
    '''

    uid, gid = conf.main_process.get_owner()

    tpl = tpl % {
        'path': UwsgiRunner.get_env_path(),
        'runtime_dir': conf.replace_placeholders('{project_runtime_dir}'),
        'user': uid,
        'group': gid,
    }

    return tpl


TEMPLATES: Dict[str, Callable] = {
    TYPE_SYSTEMD: get_tpl_systemd,
    TYPE_UPSTART: get_tpl_upstart,
}


def get_config(
        systype: str,
        *,
        conf: Union[Configuration, Section],
        conf_path: str,
        runner: str = None,
        project_name: str = None
) -> str:
    """Returns init system configuration file contents.

    :param systype: System type alias, e.g. systemd, upstart
    :param conf: Configuration/Section object.
    :param conf_path: File path to a configuration file or a command producing such a configuration.
    :param runner: Runner command to execute conf_path. Defaults to ``uwsgiconf`` runner.
    :param project_name: Project name to override.

    """
    runner = runner or f'{Finder.uwsgiconf()} run'
    conf_path = abspath(conf_path)

    if isinstance(conf, Configuration):
        conf = conf.sections[0]  # todo Maybe something more intelligent.

    tpl = dedent(TEMPLATES.get(systype)(conf=conf))

    formatted = tpl.strip().format(
        project=project_name or conf.project_name or basename(dirname(conf_path)),
        command=f'{runner} {conf_path}',
    )

    return formatted



from uwsgiconf.config import Section, Configuration, configure_uwsgi


print("this line won't print")


def get_config():

    configuration = [

        Configuration([
            Section(),
            Section('conf1_2').env('A', 'B')
        ], alias='uwsgicgf_test1'),

        Section().env('D', 'E'),
    ]
    return configuration


configure_uwsgi(get_config)



from django.db import models
from django.utils.translation import gettext_lazy as _


class Summary(models.Model):

    class Meta:
        app_label = 'uwsgify'
        managed = False
        verbose_name = _('Summary')
        verbose_name_plural = _('Summary')


class Configuration(models.Model):

    class Meta:
        app_label = 'uwsgify'
        managed = False
        verbose_name = _('Configuration')
        verbose_name_plural = _('Configuration')


class Workers(models.Model):

    class Meta:
        app_label = 'uwsgify'
        managed = False
        verbose_name = _('Workers')
        verbose_name_plural = _('Workers')


class Maintenance(models.Model):

    class Meta:
        app_label = 'uwsgify'
        managed = False
        verbose_name = _('Maintenance')
        verbose_name_plural = _('Maintenance')



from ..base import ParametrizedValue


class BalancingAlgorithm(ParametrizedValue):

    name_separator = ''


class BalancingAlgorithmWithBackup(BalancingAlgorithm):

    def __init__(self, backup_level=None):
        self.backup_level = backup_level
        super().__init__()


class WeightedRoundRobin(BalancingAlgorithmWithBackup):
    """Weighted round robin algorithm with backup support.
    The default algorithm.

    """
    name = 'wrr'


class LeastReferenceCount(BalancingAlgorithmWithBackup):
    """Least reference count algorithm with backup support."""

    name = 'lrc'


class WeightedLeastReferenceCount(BalancingAlgorithmWithBackup):
    """Weighted least reference count algorithm with backup support."""

    name = 'wlrc'


class IpHash(BalancingAlgorithmWithBackup):
    """IP hash algorithm with backup support."""

    name = 'iphash'



from ..base import OptionsGroup


class Algo(OptionsGroup):

    def __init__(self, *args, **kwargs):
        self._make_section_like()
        super().__init__(*args, **kwargs)


class AlgoSpare(Algo):
    """The **default** algorithm.

    If all workers are busy for a certain amount of time seconds then uWSGI will spawn new workers.
    When the load is gone it will begin stopping processes one at a time.

    * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#spare-cheaper-algorithm

    """
    name = 'spare'

    def set_basic_params(self, *, check_interval_overload=None):
        """
        :param int check_interval_overload: Interval (sec) to wait after all workers are busy
            before new worker spawn.

        """
        self._set('cheaper-overload', check_interval_overload)

        return self


class AlgoSpare2(Algo):
    """This algorithm is similar to ``spare``, but suitable for large scale
    by increase workers faster (before overload) and decrease them slower.

    * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#spare2-cheaper-algorithm

    """

    name = 'spare2'

    def set_basic_params(self, *, check_interval_idle=None):
        """
        :param int check_interval_idle: Decrease workers after specified idle. Default: 10.

        """
        self._set('cheaper-idle', check_interval_idle)

        return self


class AlgoQueue(Algo):
    """If the socket's listen queue has more than ``cheaper_overload`` requests
    waiting to be processed, uWSGI will spawn new workers.

    If the backlog is lower it will begin killing processes one at a time.

    * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#backlog-cheaper-algorithm

    .. note: Only available on Linux and only on TCP sockets (not UNIX domain sockets).

    """
    name = 'backlog'

    def set_basic_params(self, *, check_num_overload=None):
        """
        :param int check_num_overload: Number of backlog items in queue.

        """
        self._set('cheaper-overload', check_num_overload)

        return self._section


class AlgoBusyness(Algo):
    """Algorithm adds or removes workers based on average utilization
    for a given time period. It's goal is to keep more workers than
    the minimum needed available at any given time, so the app will always
    have capacity for new requests.

    * http://uwsgi.readthedocs.io/en/latest/Cheaper.html#busyness-cheaper-algorithm

    .. note:: Requires ``cheaper_busyness`` plugin.

    """
    name = 'busyness'
    plugin = 'cheaper_busyness'

    def set_basic_params(
            self, *, check_interval_busy=None,
            busy_max=None, busy_min=None,
            idle_cycles_max=None, idle_cycles_penalty=None,
            verbose=None):
        """
        :param int check_interval_busy: Interval (sec) to check worker busyness.

        :param int busy_max: Maximum busyness (percents). Every time the calculated busyness
            is higher than this value, uWSGI will spawn new workers. Default: 50.

        :param int busy_min: Minimum busyness (percents). If busyness is below this value,
            the app is considered in an "idle cycle" and uWSGI will start counting them.
            Once we reach needed number of idle cycles uWSGI will kill one worker. Default: 25.

        :param int idle_cycles_max: This option tells uWSGI how many idle cycles are allowed
            before stopping a worker.

        :param int idle_cycles_penalty: Number of idle cycles to add to ``idle_cycles_max``
            in case worker spawned too early. Default is 1.

        :param bool verbose: Enables debug logs for this algo.

        """
        self._set('cheaper-overload', check_interval_busy)
        self._set('cheaper-busyness-max', busy_max)
        self._set('cheaper-busyness-min', busy_min)
        self._set('cheaper-busyness-multiplier', idle_cycles_max)
        self._set('cheaper-busyness-penalty', idle_cycles_penalty)
        self._set('cheaper-busyness-verbose', verbose, cast=bool)

        return self._section

    def set_emergency_params(
            self, *, workers_step=None, idle_cycles_max=None, queue_size=None, queue_nonzero_delay=None):
        """Sets busyness algorithm emergency workers related params.

        Emergency workers could be spawned depending upon uWSGI backlog state.

        .. note:: These options are Linux only.

        :param int workers_step: Number of emergency workers to spawn. Default: 1.

        :param int idle_cycles_max: Idle cycles to reach before stopping an emergency worker. Default: 3.

        :param int queue_size: Listen queue (backlog) max size to spawn an emergency worker. Default: 33.

        :param int queue_nonzero_delay: If the request listen queue is > 0 for more than given amount of seconds
            new emergency workers will be spawned. Default: 60.

        """
        self._set('cheaper-busyness-backlog-step', workers_step)
        self._set('cheaper-busyness-backlog-multiplier', idle_cycles_max)
        self._set('cheaper-busyness-backlog-alert', queue_size)
        self._set('cheaper-busyness-backlog-nonzero', queue_nonzero_delay)

        return self


class AlgoManual(Algo):
    """Algorithm allows to adjust number of workers using Master FIFO commands.

    * http://uwsgi.readthedocs.io/en/latest/MasterFIFO.html#available-commands

    """
    name = 'manual'


class Cheapening(OptionsGroup):
    """uWSGI provides the ability to dynamically scale
    the number of running workers (adaptive process spawning) via pluggable algorithms.

    .. note:: This uses master process.

    """

    class algorithms:
        """Algorithms available to use with ``cheaper_algorithm``."""

        busyness = AlgoBusyness
        manual = AlgoManual
        queue = AlgoQueue
        spare = AlgoSpare
        spare2 = AlgoSpare2

    def set_basic_params(
            self, *, spawn_on_request=None,
            cheaper_algo=None, workers_min=None, workers_startup=None, workers_step=None):
        """
        :param bool spawn_on_request: Spawn workers only after the first request.

        :param Algo cheaper_algo: The algorithm object to be used used for adaptive process spawning.
            Default: ``spare``. See ``.algorithms``.

        :param int workers_min: Minimal workers count. Enables cheaper mode (adaptive process spawning).

            .. note:: Must be lower than max workers count.

        :param int workers_startup: The number of workers to be started when starting the application.
            After the app is started the algorithm can stop or start workers if needed.

        :param int workers_step: Number of additional processes to spawn at a time if they are needed,

        """
        self._set('cheap', spawn_on_request, cast=bool)

        if cheaper_algo:
            self._set('cheaper-algo', cheaper_algo.name)

            if cheaper_algo.plugin:
                self._section.set_plugins_params(plugins=cheaper_algo.plugin)

            cheaper_algo._contribute_to_opts(self)

        self._set('cheaper', workers_min)
        self._set('cheaper-initial', workers_startup)
        self._set('cheaper-step', workers_step)

        return self._section

    def set_memory_limits(self, *, rss_soft=None, rss_hard=None):
        """Sets worker memory limits for cheapening.

        :param int rss_soft: Don't spawn new workers if total resident memory usage
            of all workers is higher than this limit in bytes.
            
            .. warning:: This option expects memory reporting enabled:
                ``.logging.set_basic_params(memory_report=1)``

        :param int rss_hard: Try to stop workers if total workers resident memory usage
            is higher that thi limit in bytes.

        """
        self._set('cheaper-rss-limit-soft', rss_soft)           
        self._set('cheaper-rss-limit-hard', rss_hard)

        return self._section

    def print_alorithms(self):
        """Print out enabled cheaper algorithms."""

        self._set('cheaper-algo-list', True, cast=bool)

        return self._section



from typing import List, Any, Callable

from .. import uwsgi
from ..typehints import Strint
from ..utils import decode, decode_deep


class Cache:
    """Interface for uWSGI Caching subsystem.

    .. warning:: To use this helper one needs
        to configure cache(s) in uWSGI config beforehand.

        E.g.: ``section.caching.add_cache('mycache', 100)``

    """
    __slots__ = ['name', 'timeout']

    def __init__(self, name: str, *, timeout: int = None):
        """
        :param name: Cache name with optional address (if @-syntax is used).

        :param timeout: Expire timeout (seconds).
            Default: 300 (5 minutes). Use 0 to not to set a timeout (not to expire).

            .. note:: This value is ignore if cache is configured not to expire.

        """
        self.timeout = timeout or 300
        self.name = name

    def __contains__(self, key: str) -> bool:
        """Checks whether there is a value in the cache associated with the given key.

        :param key: The cache key to check.

        """
        return uwsgi.cache_exists(key, self.name)

    @property
    def keys(self) -> List[str]:
        """Returns a list of keys available in cache.

        :raises ValueError: If cache is unavailable.

        """
        return decode_deep(uwsgi.cache_keys(self.name))

    def clear(self):
        """Clears cache the cache."""
        uwsgi.cache_clear(self.name)

    def get(self, key: str, *, default: Any = None, as_int: bool = False, setter: Callable = None) -> Strint:
        """Gets a value from the cache.

        :param key: The cache key to get value for.

        :param default: Value to return if none found in cache.

        :param as_int: Return 64bit number instead of str.

        :param setter: Setter callable to automatically set cache
            value if not already cached. Required to accept a key and return
            a value that will be cached.

        """
        if as_int:
            val = uwsgi.cache_num(key, self.name)
        else:
            val = decode(uwsgi.cache_get(key, self.name))

        if val is None:

            if setter is None:
                return default

            val = setter(key)

            if val is None:
                return default

            self.set(key, val)

        return val

    __getitem__ = get

    def set(self, key: str, value: Any, *, timeout: int = None) -> bool:
        """Sets the specified key value.

        :param key:

        :param value:
            .. note:: This value will be casted to string as uWSGI cache works with strings.

        :param timeout: 0 to not to expire. Object default is used if not set.

        """
        if timeout is None:
            timeout = self.timeout

        return uwsgi.cache_set(key, str(value), timeout, self.name)

    __setitem__ = set

    def delete(self, key: str):
        """Deletes the given cached key from the cache.

        :param key: The cache key to delete.

        """
        uwsgi.cache_del(key, self.name)

    __delitem__ = delete

    def incr(self, key: str, *, delta: int = 1) -> bool:
        """Increments the specified key value by the specified value.
       
        :param key:
    
        :param delta:

        """
        return uwsgi.cache_inc(key, delta, self.timeout, self.name)

    def decr(self, key: str, *, delta: int = 1) -> bool:
        """Decrements the specified key value by the specified value.

        :param key:

        :param delta:

        """
        return uwsgi.cache_dec(key, delta, self.timeout, self.name)

    def mul(self, key: str, *, value: int = 2) -> bool:
        """Multiplies the specified key value by the specified value.

        :param key:

        :param value:

        """
        return uwsgi.cache_mul(key, value, self.timeout, self.name)

    def div(self, key: str, *, value: int = 2) -> bool:
        """Divides the specified key value by the specified value.

        :param key:

        :param value:

        """
        return uwsgi.cache_div(key, value, self.timeout, self.name)



from .. import uwsgi
from ..utils import decode


variable_set = uwsgi.set_logvar


def variable_get(name: str) -> str:
    """Return user-defined log variable contents.

    * http://uwsgi.readthedocs.io/en/latest/LogFormat.html#user-defined-logvars

    :param name:

    """
    return decode(uwsgi.get_logvar(name))


log_message = uwsgi.log

get_current_log_size = uwsgi.logsize



import os
import pickle
from calendar import timegm
from datetime import datetime, timedelta
from typing import List, Type, Callable, Dict, Union, Any, Optional

from .. import uwsgi
from ..utils import listify, get_logger, encode, decode, decode_deep

_LOG = get_logger(__name__)
_MSG_MAX_SIZE = 64 * 1024  # 64 Kb https://uwsgi-docs.readthedocs.io/en/latest/Spooler.html#spool-files

TypeTaskResult = Optional[Union['TaskResult', bool]]

_task_functions: Dict[str, Callable] = {}

spooler_task_types = {}
"""Known task types handlers will store here runtime.

SpoolerTask heirs are automatically registered 
in runtime by SpoolerTask.__init_subclass__.

"""


def _get_spoolers() -> List[str]:
    return decode_deep(listify(uwsgi.opt.get(b'spooler', [])))


def _register_task(spooler_obj: 'Spooler', spooler_cls: Type['Spooler']) -> Callable:
    """

    :param spooler_obj:
    :param spooler_cls:

    """
    def _register_task_(priority=None, postpone=None):
        # This one is returned by `_HandlerRegisterer`

        def task_decorator(func: Callable) -> Callable:
            # This one handles decoration with `Spooler.register_handler`.

            func_name = func.__name__

            _task_functions[func_name] = func

            _LOG.debug(f'Spooler. Registered function: {func_name}')

            def task_call(*args, **kwargs) -> str:
                # Redirect task (function call) into spooler.

                return spooler_cls.send_message_raw(**SpoolerFunctionCallTask.build_message(
                    spooler=spooler_obj.name if spooler_obj else None,
                    priority=priority,
                    postpone=postpone,
                    payload={
                        'func': func_name,
                        'arg': args,
                        'kwg': kwargs,
                    }
                ))

            return task_call

        return task_decorator

    return _register_task_


class _TaskRegisterer:
    # Allows task decoration using both Spooler class and object.

    def __get__(self, instance: 'Spooler', owner: Type['Spooler']):
        return _register_task(instance, owner)

    def __call__(self, *, priority: int = None, postpone: Union[datetime, timedelta] = None):
        """Decorator. Used to register a function which should be run in Spooler.

        :param priority: Number. The priority of the message. Larger - less important.

            .. warning:: This works only if you enable `order_tasks` option in `spooler.set_basic_params()`.

        :param postpone: Postpone message processing till.

        """
        # Mirrors `_register_task_` arguments for IDEs ho get proper hints.


class Spooler:
    """Gives an access to uWSGI Spooler related functions.

    .. warning:: To use this helper one needs
        to configure spooler(s) in uWSGI config beforehand.

    .. code-block:: python

        my_spooler = Spooler.get_by_basename('myspooler')

        # @Spooler.task() to  run on first available or to run on `my_spooler`:
        @my_spooler.task(postpone=timedelta(seconds=1))
        def run_me(a, b='c'):
            ...

        # Now call this function as usual and it'll run in a spooler.
        ...
        run_me('some', b='other')
        ...


    """
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name

    task = _TaskRegisterer()
    """Decorator. Used to register a function which should be run in Spooler.
    
    .. code-block:: python
       
        my_spooler = Spooler.get_by_basename('myspooler')
        
        # @Spooler.task() to  run on first available or to run on `my_spooler`:
        @my_spooler.task(postpone=timedelta(seconds=1))  
        def run_me(a, b='c'):
            ...
    
    """

    @classmethod
    def send_message_raw(
            cls,
            message: str,
            *,
            spooler: Union[str, 'Spooler'] = None,
            priority: int = None,
            postpone: Union[datetime, timedelta] = None,
            payload: Any = None
    ) -> str:
        """Sends a message to a spooler.

        :param message: Message to pass using spooler.

        :param spooler: The spooler (id or directory) to use.
            Specify the ABSOLUTE path of the spooler that has to manage this task

        :param priority: Number. The priority of the message. Larger - less important.

            .. warning:: This works only if you enable `order_tasks` option in `spooler.set_basic_params()`.

        :param postpone: Postpone message processing till.

        :param payload: Object to pickle and pass within message.

        """
        msg = {
            'msg': message,
        }

        if spooler:

            if isinstance(spooler, Spooler):
                spooler = spooler.name

            msg['spooler'] = spooler

        if priority:
            msg['priority'] = str(priority)

        if postpone:

            if isinstance(postpone, timedelta):
                postpone += datetime.utcnow()

            if isinstance(postpone, datetime):
                postpone = timegm(postpone.timetuple())

            msg['at'] = postpone

        body = {}

        if payload:
            body['payload'] = payload

        if len(repr(msg)) >= _MSG_MAX_SIZE:
            # Message is too large move into body.
            body['msg'] = msg.pop('msg')

        if body:
            msg['body'] = pickle.dumps(body)

        return uwsgi.send_to_spooler(_msg_encode(msg))

    _valid_task_results = {uwsgi.SPOOL_OK, uwsgi.SPOOL_RETRY, uwsgi.SPOOL_IGNORE}

    @classmethod
    def _process_message_raw(cls, envelope) -> int:
        decoded = _msg_decode(envelope)

        task_name = decoded['spooler_task_name']
        msg = decoded.get('msg')
        body = decoded.get('body')

        payload = None

        if body:
            body = pickle.loads(body)
            if not msg:
                msg = body.get('msg')
            payload = body.get('payload')

        task_cls = SpoolerTask

        if msg.startswith('ucfg_'):
            message_type_id = msg.replace('ucfg_', '', 1)
            task_cls = spooler_task_types.get(message_type_id, SpoolerTask)

        task = task_cls(
            name=task_name,
            message=msg,
            payload=payload,
        )

        try:
            result = task.process()

        except Exception as e:
            _LOG.exception(f"Spooler. Unhandled exception in task '{task_name}'")
            result = ResultRescheduled(exception=e)

        if result is None:
            result = ResultSkipped(result)

        elif not isinstance(result, TaskResult):
            result = ResultProcessed(result)

        return result.code_uwsgi

    @classmethod
    def get_spoolers(cls) -> List['Spooler']:
        """Returns a list of registered spoolers."""
        return [Spooler(spooler_dir) for spooler_dir in _get_spoolers()]

    @classmethod
    def get_by_basename(cls, name: str) -> Optional['Spooler']:
        """Returns spooler object for a given directory name.

        If there is more than one spooler with the same directory base name,
        the first one is returned.

        If not found `None` is returned.

        :param name: Directory base name. E.g.: 'mydir' to get spooler for '/somewhere/here/is/mydir'

        """
        spooler = None
        basename = os.path.basename

        for spooler_dir in _get_spoolers():
            if basename(spooler_dir) == name:
                spooler = Spooler(spooler_dir)
                break

        return spooler

    @classmethod
    def get_pids(cls) -> List[int]:
        """Returns a list of all spooler processes IDs."""
        return uwsgi.spooler_pids()

    @classmethod
    def set_period(cls, seconds: int) -> bool:
        """Sets how often the spooler runs.

        :param seconds:

        """
        return uwsgi.set_spooler_frequency(seconds)

    @classmethod
    def get_tasks(cls) -> List[str]:
        """Returns a list of spooler jobs (filenames in spooler directory)."""
        return uwsgi.spooler_jobs()

    @classmethod
    def read_task_file(cls, path: str) -> dict:
        """Returns a spooler task information.

        :param path: The relative or absolute path to the task to read.

        """
        return uwsgi.spooler_get_task(path) or {}


class TaskResult:
    """Represents a task processing result."""

    code_uwsgi: str = None

    def __init__(self, result: Any = None, *, exception: Exception = None):
        self.result = result
        self.exception = exception


class ResultProcessed(TaskResult):
    """Treat task as processed."""

    code_uwsgi = uwsgi.SPOOL_OK


class ResultSkipped(TaskResult):
    """Treat task as skipped (ignored)."""

    code_uwsgi = uwsgi.SPOOL_IGNORE


class ResultRescheduled(TaskResult):
    """Treat task as rescheduled (being due to retry)."""

    code_uwsgi = uwsgi.SPOOL_RETRY


class SpoolerTask:
    """Consolidates information for a spooler task."""

    mark_processed = ResultProcessed
    mark_skipped = ResultSkipped
    mark_rescheduled = ResultRescheduled

    __slots__ = ['name', 'message', 'payload']

    type_id: str = ''

    def __init__(self, name: str, message: str, payload: Any):
        self.name = name
        self.message = message
        self.payload = payload

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        spooler_task_types[cls.type_id] = cls

    def process(self) -> TypeTaskResult:
        """Processes the task.

        Supported results:
            * `None` - mark as ignored (skipped)
            * `TaskResult` - result type logic
            * exception - mark to retry
            * other - mark as processed

        """
        return True

    @classmethod
    def build_message(
            cls,
            spooler: Optional[str],
            priority: int,
            postpone: Union[datetime, timedelta],
            payload: Any = None
    ) -> dict:
        payload_ = {
            'spooler': spooler,
            'priority': priority,
            'postpone': postpone,
        }

        payload_.update(payload or {})

        msg = dict(
            message=f'ucfg_{cls.type_id}',
            spooler=spooler,
            priority=priority,
            postpone=postpone,
            payload=payload_,
        )
        return msg


class SpoolerFunctionCallTask(SpoolerTask):
    """Function call type. Allows delegating function calls to spoolers."""

    type_id = 'fcall'

    __slots__ = ['name', 'message', 'payload']

    def process(self) -> TypeTaskResult:
        payload = self.payload
        func = _task_functions[payload['func']]  # We expect an exception here if not registered.
        result = func(*payload['arg'], **payload['kwg'])
        return result


def _msg_encode(msg: dict) -> dict:
    """
    :param msg:

    """
    return {encode(k): encode(v) if isinstance(v, str) else v for k, v in msg.items()}


def _msg_decode(msg: dict) -> dict:
    """
    :param msg:

    """
    decoded = {}

    for k, v in msg.items():
        # Builtin keys are already strings. Our custom - bytes.
        k = decode(k) if isinstance(k, bytes) else k

        if k != 'body':
            # Consider body always pickled.
            v = decode(v)

        decoded[k] = v

    return decoded


uwsgi.spooler = Spooler._process_message_raw



from .. import uwsgi


connect = uwsgi.async_connect

sleep = uwsgi.async_sleep

suspend = uwsgi.suspend

get_loop_name = uwsgi.loop

wait_for_fd_read = uwsgi.wait_fd_read

wait_for_fd_write = uwsgi.wait_fd_write



from ._base import FifoCommand, Fifo


class Command(FifoCommand):

    help = 'Dump uWSGI configuration and current stats into the log'

    def run_cmd(self, fifo: Fifo, options: dict):
        fifo.cmd_stats()



from django.conf import settings


MODULE_INIT_DEFAULT = 'uwsgiinit'
"""Default name for uwsgify init modules."""


MODULE_INIT = getattr(settings, 'UWSGIFY_MODULE_INIT', MODULE_INIT_DEFAULT)
"""Name of a module with uWSGI runtime related stuff 
to automatically import from registered applications.

"""



from typing import Union

from .. import uwsgi


handshake = uwsgi.websocket_handshake


def recv(request_context=None, non_blocking: bool = False) -> bytes:
    """Receives data from websocket.

    :param request_context:
        .. note:: uWSGI 2.1+

    :param non_blocking:

    :raises IOError: If unable to receive a message.

    """
    return uwsgi.websocket_recv_nb() if non_blocking else uwsgi.websocket_recv()


def send(message: Union[str, bytes], request_context=None):
    """Sends a message to websocket.

    :param message: data to send

    :param request_context:
        .. note:: uWSGI 2.1+

    :raises IOError: If unable to send a message.

    """
    return uwsgi.websocket_send(message) if isinstance(message, str) else uwsgi.websocket_send_binary(message)



from uwsgiconf.runtime.alarms import *
from uwsgiconf.runtime.asynced import *
from uwsgiconf.runtime.control import *
from uwsgiconf.runtime.platform import uwsgi as platform
from uwsgiconf.runtime.logging import *
from uwsgiconf.runtime.request import *
from uwsgiconf.runtime.websockets import *


def test_harakiri_imposed():

    @harakiri_imposed(timeout=1)
    def doomed():
        pass

    doomed()

    with harakiri_imposed(timeout=1):
        pass



from typing import Type, Any, Union, Optional, Callable

from .utils import listify


if False:  # pragma: nocover
    from .config import Section  # noqa


class Options:
    """Options descriptor. Allows option."""

    def __init__(self, opt_type: Type['OptionsGroup']):
        """
        :param opt_type:

        """
        self.opt_type = opt_type

    def __get__(self, section: 'Section', section_cls: Type['Section']) -> Union['OptionsGroup', Type['OptionsGroup']]:
        """

        :param section:
        :param section_cls:

        """
        key = self.opt_type.__name__

        try:
            options_obj = section._options_objects.get(key)

        except AttributeError:
            # Allow easy access to option group static params:
            # Section.networking.socket_types.DEFAULT
            return self.opt_type

        if not options_obj:
            options_obj = self.opt_type(_section=section)

        section._options_objects[key] = options_obj

        return options_obj


class OptionKey:

    __slots__ = ['key']

    def __init__(self, key: str):
        self.key = key

    def swap(self, new_key: Any):

        if new_key:
            self.key = f'{new_key}'

    def __str__(self):
        return self.key

    def __eq__(self, obj):
        return self.key == obj

    def __hash__(self):
        return hash(self.key)


class OptionsGroup:
    """Introduces group of options.

    Basic group parameters may be passed to initializer
    or `set_basic_params` method.

    Usually methods setting parameters are named `set_***_params`.

    Methods ending with `_params` return section object and may be chained.

    """
    _section: 'Section' = None
    """Section this option group belongs to."""

    plugin: Union[bool, str] = False
    """Indication this option group belongs to a plugin."""

    name: str = ''
    """Name to represent the group."""

    def __init__(self, *args, **kwargs):
        if self._section is None:
            self._section = kwargs.pop('_section', None)

        self.set_basic_params(*args, **kwargs)

    def _get_name(self, *args, **kwargs) -> str:
        return self.name

    def __str__(self):
        return self._get_name()

    def __call__(self, *args, **kwargs) -> 'Section':
        """The call is translated into ``set_basic_params``` call.

        This approach is much more convenient yet IDE most probably won't
        give you a hint on what arguments are accepted.

        :param args:
        :param kwargs:

        """
        return self.set_basic_params(*args, **kwargs)

    def __eq__(self, obj):
        return f'{self}' == obj

    def __hash__(self):
        return hash(f'{self}')

    def set_basic_params(self, *args, **kwargs) -> 'Section':
        return self._section

    def _set(
            self,
            key: str,
            value: Any,
            *,
            condition: Optional[Union[bool, str]] = True,
            cast: Callable = None,
            multi: bool = False,
            plugin: str = None,
            priority: int = None
    ):
        """

        :param key: Option name

        :param value: Option value. Can be a lis if ``multi``.

        :param condition: Condition to test whether this option should be added to section.
            * True - test value is not None.

        :param cast: Value type caster.
            * bool - treat value as a flag

        :param multi: Indicate that many options can use the same name.

        :param plugin: Plugin this option exposed by. Activated automatically.

        :param priority: Option priority indicator. Options with lower numbers will come first.

        """
        key = OptionKey(key)

        def set_plugin(plugin):
            self._section.set_plugins_params(plugins=plugin)

        def handle_priority(value, *, use_list=False):

            if priority is not None:
                # Restructure options.
                opts_copy = opts.copy()
                opts.clear()

                existing_value = opts_copy.pop(key, [])

                if use_list:
                    existing_value.extend(value)
                    value = existing_value

                for pos, (item_key, item_val) in enumerate(opts_copy.items()):

                    if priority == pos:
                        opts[key] = value

                    opts[item_key] = item_val

                return True

        def handle_plugin_required(val):

            if isinstance(val, ParametrizedValue):
                key.swap(val.opt_key or key)  # todo

                if val.plugin:
                    # Automatic plugin activation.
                    set_plugin(val.plugin)

                if val._opts:
                    opts.update(val._opts)

        if condition is True:
            condition = value is not None

        if condition is None or condition:

            opts = self._section._opts

            if cast is bool:
                if value:
                    value = 'true'

                else:
                    try:
                        del opts[key]
                    except KeyError:
                        pass

                    return

            if self.plugin is True:
                # Automatic plugin activation when option from it is used.
                set_plugin(self)

            if plugin:
                set_plugin(plugin)

            if isinstance(value, tuple):  # Tuple - expect ParametrizedValue.
                list(map(handle_plugin_required, value))
                value = ' '.join(map(str, value))

            if multi:
                values = []

                # First activate plugin if required.
                for value in listify(value):
                    handle_plugin_required(value)
                    values.append(value)

                # Second: list in new option.
                if not handle_priority(values, use_list=True):
                    opts.setdefault(key, []).extend(values)

            else:
                handle_plugin_required(value)

                if not handle_priority(value):
                    opts[key] = value

    def _make_section_like(self):
        self._section = type('SectionLike', (object,), {'_opts': {}})

    def _contribute_to_opts(self, target):
        target_section = target._section
        target_section._opts.update(self._section._opts)


class ParametrizedValue(OptionsGroup):
    """Represents parametrized option value."""

    alias: str = ''
    """Alias to address this value."""

    args_joiner: str = ' '
    """Symbol to join arguments with."""

    name_separator: str = ':'
    """Separator to add after name portion."""

    name_separator_strip: bool = False
    """Strip leading and trailing name separator from the result."""

    opt_key: Optional[str] = None
    """Allows swapping default option key with custom value."""

    def __init__(self, *args):
        self.args = list(args)
        self._opts = {}
        super().__init__(_section=self)

    def __str__(self):
        args = [str(arg) for arg in self.args if arg is not None]

        result = ''

        if self.opt_key is None:
            result += self._get_name() + self.name_separator

        result += self.args_joiner.join(args)

        if self.alias:
            result = f'{self.alias} {result}'

        result = result.strip()

        if self.name_separator_strip:
            result = result.strip(self.name_separator)

        return result


class TemplatedValue(ParametrizedValue):

    tpl: str = '%s'

    def __init__(self, name: str):
        self._name = name
        super().__init__()

    def __str__(self):
        return self.tpl % self._name



from ..base import ParametrizedValue
from ..utils import KeyValue


class Pusher(ParametrizedValue):

    args_joiner = ','


class PusherSocket(Pusher):
    """Push metrics to a UDP server.

    Uses the following format: ``<metric> <type> <value>``
    ``<type>`` - is in the numeric form of metric type.

    """
    name = 'socket'
    plugin = 'stats_pusher_socket'

    def __init__(self, address, *, prefix=None):
        """
        :param str address:

        :param str prefix: Arbitrary prefix to differentiate sender.

        """
        super().__init__(address, prefix)


class PusherRrdtool(Pusher):
    """This will store an rrd file for each metric in the specified directory.

    Each rrd file has a single data source named "metric".

    """
    name = 'rrdtool'
    plugin = 'rrdtool'

    def __init__(self, target_dir, *, library=None, push_interval=None):
        """
        :param str target_dir: Directory to store rrd files into.

        :param str library: Set the name of rrd library. Default: librrd.so.

        :param int push_interval: Set push frequency.

        """
        super().__init__(target_dir)

        self._set('rrdtool-freq', push_interval)
        self._set('rrdtool-lib', library)


class PusherStatsd(Pusher):
    """Push metrics to a statsd server."""

    name = 'statsd'
    plugin = 'stats_pusher_statsd'

    def __init__(self, address, *, prefix=None, no_workers=None, all_gauges=None):
        """
        :param str address:

        :param str prefix: Arbitrary prefix to differentiate sender.

        :param bool no_workers: Disable generation of single worker metrics.

        :param bool all_gauges: Push all metrics to statsd as gauges.
        """
        super().__init__(address, prefix)

        self._set('statsd-no-workers', no_workers, cast=bool)
        self._set('statsd-all-gauges', all_gauges, cast=bool)


class PusherCarbon(Pusher):
    """Push metrics to a Carbon server of Graphite.

    Metric node format: ``<node_root>.hostname.<node_realm>.metrics_data``.

    * http://uwsgi.readthedocs.io/en/latest/Carbon.html
    * http://uwsgi.readthedocs.io/en/latest/tutorials/GraphiteAndMetrics.html

    """
    name = 'carbon'
    plugin = 'carbon'
    opt_key = name

    def __init__(
            self, address, *, node_realm=None, node_root=None, push_interval=None, idle_avg_source=None,
            use_metrics=None, no_workers=None, timeout=None, retries=None, retries_delay=None,
            hostname_dots_replacer=None):
        """
        :param str|list[str] address: Host and port. Example: 127.0.0.1:2004

        :param str node_realm: Set carbon metrics realm node.

        :param str node_root: Set carbon metrics root node. Default: uwsgi.

        :param int push_interval: Set carbon push frequency in seconds. Default: 60.

        :param bool no_workers: Disable generation of single worker metrics.

        :param str idle_avg_source: Average values source during idle period (no requests).

            Variants:
                * last (default)
                * zero
                * none

        :param bool use_metrics: Don't compute all statistics, use metrics subsystem data
            instead.

            .. warning:: Key names of built-in stats are different from those of metrics system.

        :param int timeout: Set carbon connection timeout in seconds. Default: 3.

        :param int retries: Set maximum number of retries in case of connection errors. Default: 1.

        :param int retries_delay: Set connection retry delay in seconds. Default: 7.

        :param str hostname_dots_replacer: Set char to use as a replacement for
            dots in hostname in `<node_root>.hostname.<node_realm>.metrics_data``

            This affects Graphite aggregation mechanics.

            .. note:: Dots are not replaced by default.

        """
        super().__init__(address)

        self._set('carbon-id', node_realm)
        self._set('carbon-root', node_root)
        self._set('carbon-freq', push_interval)
        self._set('carbon-idle-avg', idle_avg_source)
        self._set('carbon-use-metrics', use_metrics, cast=bool)
        self._set('carbon-no-workers', no_workers, cast=bool)
        self._set('carbon-timeout', timeout)
        self._set('carbon-max-retry', retries)
        self._set('carbon-retry-delay', retries_delay)
        self._set('carbon-hostname-dots', hostname_dots_replacer)

        if not address.split(':')[0].replace('.', '').isdigit():
            self._set('carbon-name-resolve', True, cast=bool)


class PusherZabbix(Pusher):
    """Push metrics to a zabbix server."""

    name = 'zabbix'
    plugin = 'zabbix'

    def __init__(self, address, *, prefix=None, template=None):
        """
        :param str address:

        :param str prefix: Arbitrary prefix to differentiate sender.

        :param str template: Print (or store to a file) the zabbix template
            for the current metrics setup.

        """
        super().__init__(address, prefix)

        self._set('zabbix-template', template)


class PusherMongo(Pusher):
    """Push statistics (as JSON) the the specified MongoDB database."""

    name = 'mongodb'
    plugin = 'stats_pusher_mongodb'

    def __init__(self, *, address=None, collection=None, push_interval=None):
        """
        :param str address: Default: 127.0.0.1:27017

        :param str collection: MongoDB colection to write into. Default: uwsgi.statistics

        :param int push_interval: Write interval in seconds.
        """
        value = KeyValue(locals(), aliases={'push_interval': 'freq'})

        super().__init__(value)


class PusherFile(Pusher):
    """Stores stats JSON into a file.

    .. note:: Mainly for demonstration purposes.

    """
    name = 'file'
    plugin = 'stats_pusher_file'

    def __init__(self, fpath=None, *, separator=None, push_interval=None):
        """
        :param str fpath: File path. Default: uwsgi.stats

        :param str separator: New entry separator. Default: \n\n

        :param int push_interval: Write interval in seconds.
        """
        value = KeyValue(locals(), aliases={'fpath': 'path', 'push_interval': 'freq'})

        super().__init__(value)



from ..base import ParametrizedValue


class SubjectBuiltin:

    name = None

    def __init__(self, regexp):
        self.regexp = regexp


class SubjectCustom(ParametrizedValue):
    """Represents a routing subject that supports various check."""

    args_joiner = ';'

    def __init__(self, subject, *, negate=False):
        """
        :param Var|str subject: Handwritten subject or a Var heir representing it.

        :param bool negate: Use to negate subject for rule.
            .. note:: You can also use tilde (~) instead of this argument for nagation.

        """
        self._subject = subject
        self.negate = negate
        super().__init__()

    def __invert__(self):
        self.negate = True
        return self

    def _setup(self, name, arg=None):
        self.name = f'{name}'
        self.args = [self._subject, arg]
        return self

    def exists(self):
        """Check if the subject exists in the filesystem."""
        return self._setup('exists')

    def isfile(self):
        """Check if the subject is a file."""
        return self._setup('isfile')

    def isdir(self):
        """Check if the subject is a directory."""
        return self._setup('isdir')

    def islink(self):
        """Check if the subject is a link."""
        return self._setup('islink')

    def isexec(self):
        """Check if the subject is an executable file."""
        return self._setup('isexec')

    def islord(self):
        """Check if the subject is a Legion Lord."""
        return self._setup('lord')

    def contains_ipv4(self):
        """Check if the subject is ip v4."""
        return self._setup('ipv4in')

    def contains_ipv6(self):
        """Check if the subject is ip v6."""
        return self._setup('ipv6in')

    def eq(self, val):
        """Check if the subject is equal to the specified pattern."""
        return self._setup('eq', val)

    def ge(self, val):
        """Check if the subject is greater than or equal to the specified pattern."""
        return self._setup('ishigherequal', val)

    def le(self, val):
        """Check if the subject is less than or equal to the specified pattern."""
        return self._setup('islowerequal', val)

    def gt(self, val):
        """Check if the subject is greater than the specified pattern."""
        return self._setup('ishigher', val)

    def lt(self, val):
        """Check if the subject is less than the specified pattern."""
        return self._setup('islower', val)

    def startswith(self, val):
        """Check if the subject starts with the specified pattern."""
        return self._setup('startswith', val)

    def endswith(self, val):
        """Check if the subject ends with the specified pattern."""
        return self._setup('endswith', val)

    def matches(self, regexp):
        """Check if the subject matches the specified regexp."""
        return self._setup('regexp', regexp)

    def isempty(self):
        """Check if the subject is empty."""
        return self._setup('empty')

    def contains(self, val):
        """Check if the subject contains the specified pattern."""
        return self._setup('contains', val)


class SubjectPathInfo(SubjectBuiltin):
    """Default subject, maps to PATH_INFO."""

    name = ''


class SubjectRequestUri(SubjectBuiltin):
    """Checks REQUEST_URI for a value."""

    name = 'uri'


class SubjectQueryString(SubjectBuiltin):
    """Checks QUERY_STRING for a value."""

    name = 'qs'


class SubjectRemoteAddr(SubjectBuiltin):
    """Checks REMOTE_ADDR for a value."""

    name = 'remote-addr'


class SubjectRemoteUser(SubjectBuiltin):
    """Checks REMOTE_USER for a value."""

    name = 'remote-user'


class SubjectHttpHost(SubjectBuiltin):
    """Checks HTTP_HOST for a value."""

    name = 'host'


class SubjectHttpReferer(SubjectBuiltin):
    """Checks HTTP_REFERER for a value."""

    name = 'referer'


class SubjectHttpUserAgent(SubjectBuiltin):
    """Checks HTTP_USER_AGENT for a value."""

    name = 'user-agent'


class SubjectStatus(SubjectBuiltin):
    """Checks HTTP response status code.

    .. warning:: Not available in the request chain.

    """
    name = 'status'
